from landscape.lib import md5crypt
from landscape.user.management import UserManagement, UserManagementError
from landscape.user.tests.helpers import FakeUserProvider
from landscape.user.provider import UserNotFoundError, GroupNotFoundError
from landscape.tests.helpers import LandscapeTest, MockPopen


def guess_password(generated_password, plaintext_password):
    salt = generated_password[len("$1$"):generated_password.rfind("$")]
    crypted = md5crypt.md5crypt(plaintext_password, salt)
    return crypted


class UserWriteTest(LandscapeTest):

    def setUp(self):
        LandscapeTest.setUp(self)
        self.shadow_file = self.makeFile("""\
jdoe:$1$xFlQvTqe$cBtrNEDOIKMy/BuJoUdeG0:13348:0:99999:7:::
psmith:!:13348:0:99999:7:::
sbarnes:$1$q7sz09uw$q.A3526M/SHu8vUb.Jo1A/:13349:0:99999:7:::
""")

    def test_add_user(self):
        """L{UserManagement.add_user} should use C{adduser} to add users."""
        groups = [("users", "x", 1001, [])]
        provider = FakeUserProvider(groups=groups, popen=MockPopen(""))
        management = UserManagement(provider=provider)
        management.add_user("jdoe", "John Doe", "password", False, "users",
                            "Room 101", "+123456", None)
        self.assertEqual(len(provider.popen.popen_inputs), 2)
        self.assertEqual(provider.popen.popen_inputs[0],
                         ["adduser", "jdoe", "--disabled-password",
                          "--gecos", "John Doe,Room 101,+123456,",
                          "--gid", "1001"])

        usermod = provider.popen.popen_inputs[1]
        self.assertEqual(len(usermod), 4, usermod)
        password = guess_password(usermod[2], "password")
        self.assertEqual(usermod, ["usermod", "-p", password, "jdoe"])

    def test_add_user_error(self):
        """
        L{UserManagement.add_user} should raise an L{UserManagementError} if
        C{adduser} fails.
        """
        provider = FakeUserProvider(popen=MockPopen("", return_codes=[1, 0]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserManagementError, management.add_user,
                          "jdoe", u"John Doe", "password", False, None, None,
                          None, None)

    def test_change_password_error(self):
        """
        L{UserManagement.add_user} should raise an L{UserManagementError} if
        C{usermod} fails.
        """
        provider = FakeUserProvider(popen=MockPopen("", return_codes=[0, 1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserManagementError, management.add_user,
                          "jdoe", u"John Doe", "password", False, None, None,
                          None, None)

    def test_expire_password_error(self):
        """
        L{UserManagement.add_user} should raise an L{UserManagementError} if
        C{passwd} fails.
        """
        provider = FakeUserProvider(
            popen=MockPopen("", return_codes=[0, 0, 1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserManagementError, management.add_user,
                          "jdoe", u"John Doe", "password", True, None, None,
                          None, None)

    def test_set_password(self):
        """
        L{UserManagement.set_password} should use C{usermod} to change
        a user's password.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", password="password")

        self.assertEqual(len(provider.popen.popen_inputs), 1)
        password = provider.popen.popen_inputs[0][2]
        password = guess_password(password, "password")
        self.assertEqual(provider.popen.popen_inputs,
                         [["usermod", "-p", password, "jdoe"]])

    def test_set_password_with_system_user(self):
        """
        L{UserManagement.set_password} should allow us to edit system
        users.
        """
        data = [("root", "x", 0, 0, ",,,,", "/home/root", "/bin/zsh")]
        provider = FakeUserProvider(users=data,
                                    shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("root", password="password")
        self.assertEqual(len(provider.popen.popen_inputs), 1)
        password = provider.popen.popen_inputs[0][2]
        password = guess_password(password, "password")
        self.assertEqual(provider.popen.popen_inputs,
                         [["usermod", "-p", password, "root"]])

    def test_set_password_unicode(self):
        """
        Make sure passing unicode as username and password doesn't
        change things much (note that using something that's
        non-ASCII-encodable still probably won't work).
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", password=u"password")

        self.assertEqual(len(provider.popen.popen_inputs), 1)
        password = provider.popen.popen_inputs[0][2]
        password = guess_password(password, "password")
        self.assertEqual(provider.popen.popen_inputs,
                         [["usermod", "-p", password, "jdoe"]])

    def test_set_name(self):
        """
        L{UserManagement.set_user_details} should use C{chfn} to
        change a user's name.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", name="JD")

        self.assertEqual(len(provider.popen.popen_inputs), 1)
        self.assertEqual(provider.popen.popen_inputs,
                         [["chfn", "-f", "JD", "jdoe"]])

    def test_set_location(self):
        """
        L{UserManagement.set_user_details} should use C{chfn} to
        change a user's location.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", location="Everywhere")

        self.assertEqual(len(provider.popen.popen_inputs), 1)
        self.assertEqual(provider.popen.popen_inputs,
                         [["chfn", "-r", "Everywhere", "jdoe"]])

    def test_clear_user_location(self):
        """
        L{UserManagement.set_user_details} should use C{chfn} to
        change a user's location.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,Room 101,,,", "/home/jdoe",
                 "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", location="")
        self.assertEqual(len(provider.popen.popen_inputs), 1)
        self.assertEqual(provider.popen.popen_inputs,
                         [["chfn", "-r", "", "jdoe"]])

    def test_clear_telephone_numbers(self):
        """
        L{UserManagement.set_user_details} should use C{chfn} to
        change a user's telephone numbers.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,+123456,+123456", "/home/jdoe",
                 "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", home_number="", work_number="")
        self.assertEqual(len(provider.popen.popen_inputs), 1)
        self.assertEqual(provider.popen.popen_inputs,
                         [["chfn", "-w", "", "-h", "", "jdoe"]])

    def test_set_user_details_fails(self):
        """
        L{UserManagement.set_user_details} should raise an
        L{EditUserError} if C{chfn} fails.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("", return_codes=[1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.set_user_details, 1000,
                          name="John Doe")

    def test_contact_details_in_general(self):
        """
        L{UserManagement.set_user_details} should use C{chfn} to
        change a user's contact details.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        location = u"Everywhere"
        work_number = u"1-800-123-4567"
        home_number = u"764-4321"
        management.set_user_details("jdoe", location=location,
                                    work_number=work_number,
                                    home_number=home_number)

        self.assertEqual(len(provider.popen.popen_inputs), 1)
        self.assertEqual(provider.popen.popen_inputs,
                         [["chfn", "-r", location, "-w", work_number,
                           "-h", home_number, "jdoe"]])

    def test_set_user_details_with_unknown_username(self):
        """
        L{UserManagement.set_user_details} should raise a
        L{UserManagementError} if the user being edited doesn't exist.
        """
        provider = FakeUserProvider(popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.set_user_details,
                          "kevin", name=u"John Doe")

    def test_set_primary_group(self):
        """
        L{UserManagement.set_set_user_details} should use C{usermod} to change
        the user's primary group.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("users", "x", 1001, [])]
        provider = FakeUserProvider(users=data, groups=groups,
                                    shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))

        management = UserManagement(provider=provider)
        management.set_user_details("jdoe", primary_group_name="users")

        self.assertEqual(provider.popen.popen_inputs,
                         [["usermod", "-g", "1001", "jdoe"]])

    def test_set_primary_group_unknown_group(self):
        """
        L{UserManagement.set_user_details should use C{usermod} to change the
        user's primary group, in the event that we have an invalid group, we
        should raise a UserManagement error.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("staff", "x", 1001, [])]
        provider = FakeUserProvider(users=data, groups=groups,
                                    shadow_file=self.shadow_file,
                                    popen=MockPopen("group id 1002 unknown",
                                    return_codes=[1]))
        management = UserManagement(provider=provider)
        self.assertRaises(GroupNotFoundError, management.set_user_details,
                          "jdoe", primary_group_name="unknown")

    def test_lock_user(self):
        """L{UserManagement.lock_user} should use C{usermod} to lock users."""
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.lock_user("jdoe")
        self.assertEqual(provider.popen.popen_inputs,
                         [["usermod", "-L", "jdoe"]])

    def test_lock_user_fails(self):
        """
        L{UserManagement.lock_user} should raise a L{UserManagementError} if
        a C{usermod} fails.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("", [1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.lock_user, 1000)

    def test_lock_user_with_unknown_uid(self):
        """
        L{UserManagement.lock_user} should raise a L{UserManagementError}
        if the user being removed doesn't exist.
        """
        provider = FakeUserProvider(popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.lock_user, 1000)

    def test_unlock_user(self):
        """
        L{UserManagement.unlock_user} should use C{usermod} to unlock
        users.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.unlock_user("jdoe")
        self.assertEqual(provider.popen.popen_inputs,
                         [["usermod", "-U", "jdoe"]])

    def test_unlock_user_fails(self):
        """
        L{UserManagement.unlock_user} should raise an
        L{UserManagementError} if a C{usermod} fails.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=MockPopen("", [1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.unlock_user, 1000)

    def test_unlock_user_with_unknown_uid(self):
        """
        L{UserManagement.unlock_user} should raise a
        L{UserManagementError} if the user being removed doesn't exist.
        """
        provider = FakeUserProvider(popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.unlock_user, 1000)

    def test_remove_user(self):
        """
        L{UserManagement.remove_user} should use C{deluser} to remove
        users.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        popen = MockPopen("Removing user `jdoe'...\r\ndone.")
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=popen)
        management = UserManagement(provider=provider)
        management.remove_user("jdoe")
        self.assertEqual(popen.popen_inputs, [["deluser", "jdoe"]])

    def test_remove_user_with_unknown_username(self):
        """
        L{UserManagement.remove_user} should raise a
        L{UserManagementError} if the user being removed doesn't exist.
        """
        provider = FakeUserProvider(popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.remove_user, "smith")

    def test_remove_user_fails(self):
        """
        L{UserManagement.remove_user} should raise a
        L{UserManagementError} if the user can't be removed.
        """
        self.log_helper.ignore_errors(UserNotFoundError)
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        popen = MockPopen("/usr/sbin/deluser: Only root may remove a user or "
                          "group from the system.", [1])
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=popen)
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.remove_user, "smith")

    def test_remove_user_and_home(self):
        """
        L{UserManagement.remove_user} should use C{deluser} to remove
        the contents of a user's home directory.
        """
        data = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        popen = MockPopen("Removing user `jdoe`...\r\ndone.", [0])
        provider = FakeUserProvider(users=data, shadow_file=self.shadow_file,
                                    popen=popen)
        management = UserManagement(provider=provider)
        management.remove_user("jdoe", delete_home=True)
        self.assertEqual(popen.popen_inputs,
                         [["deluser", "jdoe", "--remove-home"]])


class GroupWriteTest(LandscapeTest):

    def setUp(self):
        LandscapeTest.setUp(self)
        self.shadow_file = self.makeFile("""\
jdoe:$1$xFlQvTqe$cBtrNEDOIKMy/BuJoUdeG0:13348:0:99999:7:::
psmith:!:13348:0:99999:7:::
sbarnes:$1$q7sz09uw$q.A3526M/SHu8vUb.Jo1A/:13349:0:99999:7:::
""")

    def test_add_group(self):
        """
        L{UserManagement.add_group} should use the system tool
        C{addgroup} to create groups.
        """
        provider = FakeUserProvider(popen=MockPopen("Result"))
        management = UserManagement(provider=provider)
        result = management.add_group("webdev")
        self.assertEqual(provider.popen.popen_inputs,
                         [["addgroup", "webdev"]])
        self.assertEqual(result, "Result")

    def test_add_group_handles_errors(self):
        """
        If the system tool C{addgroup} returns a non-0 exit code,
        L{UserManagement.add_group} should raise an L{UserManagementError}.
        """
        provider = FakeUserProvider(popen=MockPopen("Error Result", [1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserManagementError, management.add_group, "kaboom")

    def test_set_group_details(self):
        """
        L{UserManagement.set_group_details} should use C{groupmode} to
        change a group's name.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(
            users=users, shadow_file=self.shadow_file,
            groups=groups, popen=MockPopen("no output"))
        management = UserManagement(provider=provider)
        management.set_group_details("bizdev", "sales")

        self.assertEqual(provider.popen.popen_inputs,
                         [["groupmod", "-n", "sales", "bizdev"]])

    def test_set_group_details_with_unknown_groupname(self):
        """
        L{UserManagement.set_group_details} should raise a
        L{UserManagementError} if the group being updated doesn't exist.
        """
        provider = FakeUserProvider(popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(GroupNotFoundError, management.set_group_details,
                          "sales", u"newsales")

    def test_set_group_details_fails(self):
        """
        L{UserManagement.set_group_details} should raise a
        L{UserManagementError} if the group can't be renamed.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 1001, [])]
        popen = MockPopen("groupmod: sales is not a unique name", [1])
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups, popen=popen)
        management = UserManagement(provider=provider)
        self.assertRaises(UserManagementError, management.set_group_details,
                          "bizdev", u"sales")

    def test_add_member(self):
        """
        L{UserManagement.add_group_member} should use the system tool
        C{gpasswd} via the process factory to add a member to a group.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups,
                                    popen=MockPopen(
                                    "Removing user jdoe from group bizdev"))
        management = UserManagement(provider=provider)

        output = management.add_group_member("jdoe", "bizdev")
        self.assertEqual(provider.popen.popen_inputs,
                         [["gpasswd", "-a", "jdoe", "bizdev"]])
        self.assertEqual(output, "Removing user jdoe from group bizdev")

    def test_add_member_with_unknown_groupname(self):
        """
        L{UserManagement.add_group_member} should raise a
        L{UserManagementError} if the group to add the member to doesn't
        exist.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(GroupNotFoundError, management.add_group_member,
                          "jdoe", "bizdev")

    def test_add_member_with_unknown_username(self):
        """
        L{UserManagement.add_group_member} should raise a
        L{UserManagementError} if the user being associated doesn't
        exist.
        """
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(groups=groups, popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.add_group_member,
                          "bizdev", "smith")

    def test_add_member_failure(self):
        """
        If adding a member to a group fails,
        L{UserManagement.add_group_member} should raise an
        L{UserManagementError}.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups,
                                    popen=MockPopen("no output", [1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.add_group_member,
                          1000, 1001)

    def test_remove_member(self):
        """
        L{UserManagement.remove_group_member} should use the system
        tool C{gpasswd} via the process factory to remove a member
        from a group.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups,
                                    popen=MockPopen(
                                    "Removing user jdoe from group bizdev"))
        management = UserManagement(provider=provider)
        output = management.remove_group_member("jdoe", "bizdev")
        self.assertEqual(provider.popen.popen_inputs,
                         [["gpasswd", "-d", "jdoe", "bizdev"]])
        self.assertEqual(output, "Removing user jdoe from group bizdev")

    def test_remove_member_with_unknown_groupname(self):
        """
        L{UserManagement.remove_group_member} should raise a
        L{UserManagementError} if the group to remove the member to
        doesn't exist.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    popen=MockPopen("", return_codes=[2]))
        management = UserManagement(provider=provider)
        self.assertRaises(GroupNotFoundError, management.remove_group_member,
                          "jdoe", "bizdev")

    def test_remove_member_with_unknown_username(self):
        """
        L{UserManagement.remove_group_member} should raise a
        L{UserManagementError} if the user being associated doesn't
        exist.
        """
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(groups=groups,
                                    popen=MockPopen("", return_codes=[4]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserNotFoundError, management.remove_group_member,
                          "jdoe", "bizdev")

    def test_remove_member_failure(self):
        """
        If removing a member from a group fails,
        L{UserManagement.remove_group_member} should raise a
        L{UserManagementError}.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 1001, [])]
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups,
                                    popen=MockPopen("no output", [1]))
        management = UserManagement(provider=provider)
        self.assertRaises(UserManagementError,
                          management.remove_group_member, "jdoe", "bizdev")

    def test_remove_group(self):
        """
        L{UserManagement.remove_group} should use C{groupdel} to
        remove groups.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 50, [])]
        popen = MockPopen("Removing group `bizdev'...\r\ndone.")
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups, popen=popen)
        management = UserManagement(provider=provider)
        management.remove_group("bizdev")
        self.assertEqual(provider.popen.popen_inputs,
                         [["groupdel", "bizdev"]])

    def test_remove_group_with_unknown_groupname(self):
        """
        L{UserManagement.remove_group} should raise a
        L{GroupMissingError} if the group being removed doesn't exist.
        """
        provider = FakeUserProvider(popen=MockPopen(""))
        management = UserManagement(provider=provider)
        self.assertRaises(
            GroupNotFoundError, management.remove_group, "ubuntu")

    def test_remove_group_fails(self):
        """
        L{UserManagement.remove_user} should raise a
        L{RemoveUserError} if the user can't be removed.
        """
        users = [("jdoe", "x", 1000, 1000, "JD,,,,", "/home/jdoe", "/bin/zsh")]
        groups = [("bizdev", "x", 50, [])]
        popen = MockPopen("/usr/sbin/deluser: Only root may remove a user or "
                          "group from the system.", [1])
        provider = FakeUserProvider(users=users, shadow_file=self.shadow_file,
                                    groups=groups, popen=popen)
        management = UserManagement(provider=provider)
        self.assertRaises(
            GroupNotFoundError, management.remove_group, "ubuntu")



"""A utility module which has FD-related functions.

This module mostly exists for L{clean_fds}, so it can be imported without
accidentally getting a reactor or something else that might create a critical
file descriptor.
"""

import os
import resource


def clean_fds():
    """Close all non-stdio file descriptors.

    This should be called at the beginning of a program to avoid inheriting any
    unwanted file descriptors from the invoking process.  Unfortunately, this
    is really common in unix!
    """
    rlimit_nofile = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    total_descriptors = min(4096, rlimit_nofile)
    for fd in range(3, total_descriptors):
        try:
            os.close(fd)
        except OSError:
            pass



"""
Network introspection utilities using ioctl and the /proc filesystem.
"""
import os

from landscape.lib.fs import read_file


def get_vm_info(root_path="/"):
    """
    Return a string with the virtualization type if it's known, an empty string
    otherwise.

    It loops through some possible configurations and return a string with
    the name of the technology being used or None if there's no match
    """
    def join_root_path(path):
        return os.path.join(root_path, path)

    xen_paths = ["proc/sys/xen", "proc/xen"]
    xen_paths = map(join_root_path, xen_paths)
    vz_path = join_root_path("proc/vz")

    if os.path.exists(vz_path):
        return "openvz"
    elif filter(os.path.exists, xen_paths):
        return "xen"

    # /sys/bus/xen exists on most machines, but only virtual machines have
    # devices
    sys_xen_path = join_root_path("sys/bus/xen/devices")
    if os.path.isdir(sys_xen_path) and os.listdir(sys_xen_path):
        return "xen"

    sys_vendor_path = join_root_path("sys/class/dmi/id/sys_vendor")
    if not os.path.exists(sys_vendor_path):
        # Some virtualised CPU architectures indicate this via cpuinfo
        cpuinfo_path = join_root_path("proc/cpuinfo")
        if os.path.exists(cpuinfo_path):
            cpuinfo = read_file(cpuinfo_path)
            if "qemu" in cpuinfo:
                return "kvm"
        return ""

    vendor = read_file(sys_vendor_path)
    content_vendors_map = (
        ("VMware, Inc.", "vmware"),
        ("Microsoft Corporation", "hyperv"),
        ("Bochs", "kvm"),
        ("OpenStack", "kvm"),
        ("innotek GmbH", "virtualbox"))
    for name, vm_type in content_vendors_map:
        if name in vendor:
            return vm_type

    return ""


def get_container_info(path="/run/container_type"):
    """
    Return a string with the type of container the client is running in, if
    any, an empty string otherwise.
    """
    return read_file(path).strip() if os.path.exists(path) else ""



import os

from twisted.internet.defer import Deferred


from landscape.manager.haservice import HAService
from landscape.manager.plugin import SUCCEEDED, FAILED
from landscape.tests.helpers import LandscapeTest, ManagerHelper
from landscape.tests.mocker import ANY


class HAServiceTests(LandscapeTest):
    helpers = [ManagerHelper]

    def setUp(self):
        super(HAServiceTests, self).setUp()
        self.ha_service = HAService()
        self.ha_service.JUJU_UNITS_BASE = self.makeDir()

        self.unit_name = "my-service/9"
        self.unit_path = "unit-" + self.unit_name.replace("/", "-")

        self.scripts_dir = os.path.join(
            self.ha_service.JUJU_UNITS_BASE, self.unit_path, "charm/scripts")
        self.health_check_d = os.path.join(
            self.scripts_dir, self.ha_service.HEALTH_SCRIPTS_DIR)
        # create entire dir path
        os.makedirs(self.health_check_d)

        self.manager.add(self.ha_service)

        cluster_online = self.makeFile(
            content="#!/bin/bash\nexit 0",
            basename="add_to_cluster", dirname=self.scripts_dir)
        os.chmod(cluster_online, 0755)
        cluster_standby = self.makeFile(
            content="#!/bin/bash\nexit 0",
            basename="remove_from_cluster", dirname=self.scripts_dir)
        os.chmod(cluster_standby, 0755)

        service = self.broker_service
        service.message_store.set_accepted_types(["operation-result"])

    def test_invalid_server_service_state_request(self):
        """
        When the landscape server requests a C{service-state} other than
        'online' or 'standby' the client responds with the appropriate error.
        """
        logging_mock = self.mocker.replace("logging.error")
        logging_mock("Invalid cluster participation state requested BOGUS.")
        self.mocker.replay()

        self.manager.dispatch_message(
            {"type": "change-ha-service", "service-name": "my-service",
             "unit-name": self.unit_name, "service-state": "BOGUS",
             "operation-id": 1})

        service = self.broker_service
        self.assertMessages(
            service.message_store.get_pending_messages(),
            [{"type": "operation-result", "result-text":
              u"Invalid cluster participation state requested BOGUS.",
              "status": FAILED, "operation-id": 1}])

    def test_not_a_juju_computer(self):
        """
        When not a juju charmed computer, L{HAService} reponds with an error
        due to missing JUJU_UNITS_BASE dir.
        """
        self.ha_service.JUJU_UNITS_BASE = "/I/don't/exist"

        logging_mock = self.mocker.replace("logging.error")
        logging_mock("This computer is not deployed with juju. "
                     "Changing high-availability service not supported.")
        self.mocker.replay()

        self.manager.dispatch_message(
            {"type": "change-ha-service", "service-name": "my-service",
             "unit-name": self.unit_name,
             "service-state": self.ha_service.STATE_STANDBY,
             "operation-id": 1})

        service = self.broker_service
        self.assertMessages(
            service.message_store.get_pending_messages(),
            [{"type": "operation-result", "result-text":
              u"This computer is not deployed with juju. Changing "
              u"high-availability service not supported.",
              "status": FAILED, "operation-id": 1}])

    def test_incorrect_juju_unit(self):
        """
        When not the specific juju charmed computer, L{HAService} reponds
        with an error due to missing the JUJU_UNITS_BASE/$JUJU_UNIT dir.
        """
        logging_mock = self.mocker.replace("logging.error")
        logging_mock("This computer is not juju unit some-other-service-0. "
                     "Unable to modify high-availability services.")
        self.mocker.replay()

        self.manager.dispatch_message(
            {"type": "change-ha-service", "service-name": "some-other-service",
             "unit-name": "some-other-service-0", "service-state": "standby",
             "operation-id": 1})

        service = self.broker_service
        self.assertMessages(
            service.message_store.get_pending_messages(),
            [{"type": "operation-result", "result-text":
              u"This computer is not juju unit some-other-service-0. "
              u"Unable to modify high-availability services.",
              "status": FAILED, "operation-id": 1}])

    def test_wb_no_health_check_directory(self):
        """
        When unable to find a valid C{HEALTH_CHECK_DIR}, L{HAService} will
        succeed but log an informational message.
        """
        self.ha_service.HEALTH_SCRIPTS_DIR = "I/don't/exist"

        def should_not_be_called(result):
            self.fail(
                "_run_health_checks failed on absent health check directory.")

        def check_success_result(result):
            self.assertEqual(
                result,
                "Skipping juju charm health checks. No scripts at "
                "%s/I/don't/exist." % self.scripts_dir)

        result = self.ha_service._run_health_checks(self.scripts_dir)
        result.addCallbacks(check_success_result, should_not_be_called)

    def test_wb_no_health_check_scripts(self):
        """
        When C{HEALTH_CHECK_DIR} exists but, no scripts exist, L{HAService}
        will log an informational message, but succeed.
        """
        # In setup we created a health check directory but placed no health
        # scripts in it.
        def should_not_be_called(result):
            self.fail(
                "_run_health_checks failed on empty health check directory.")

        def check_success_result(result):
            self.assertEqual(
                result,
                "Skipping juju charm health checks. No scripts at "
                "%s/%s." %
                (self.scripts_dir, self.ha_service.HEALTH_SCRIPTS_DIR))

        result = self.ha_service._run_health_checks(self.scripts_dir)
        result.addCallbacks(check_success_result, should_not_be_called)

    def test_wb_failed_health_script(self):
        """
        L{HAService} runs all health check scripts found in the
        C{HEALTH_CHECK_DIR}. If any script fails, L{HAService} will return a
        deferred L{fail}.
        """

        def expected_failure(result):
            self.assertEqual(
                str(result.value),
                "Failed charm script: %s/%s/my-health-script-2 "
                "exited with return code 1." %
                (self.scripts_dir, self.ha_service.HEALTH_SCRIPTS_DIR))

        def check_success_result(result):
            self.fail(
                "_run_health_checks succeded despite a failed health script.")

        for number in [1, 2, 3]:
            script_path = (
                "%s/my-health-script-%d" % (self.health_check_d, number))
            health_script = file(script_path, "w")
            if number == 2:
                health_script.write("#!/bin/bash\nexit 1")
            else:
                health_script.write("#!/bin/bash\nexit 0")
            health_script.close()
            os.chmod(script_path, 0755)

        result = self.ha_service._run_health_checks(self.scripts_dir)
        result.addCallbacks(check_success_result, expected_failure)
        return result

    def test_missing_cluster_standby_or_cluster_online_scripts(self):
        """
        When no cluster status change scripts are delivered by the charm,
        L{HAService} will still return a L{succeeded}.
        C{HEALTH_CHECK_DIR}. If any script fails, L{HAService} will return a
        deferred L{fail}.
        """

        def should_not_be_called(result):
            self.fail(
                "_change_cluster_participation failed on absent charm script.")

        def check_success_result(result):
            self.assertEqual(
                result,
                "This computer is always a participant in its high-availabilty"
                " cluster. No juju charm cluster settings changed.")

        self.ha_service.CLUSTER_ONLINE = "I/don't/exist"
        self.ha_service.CLUSTER_STANDBY = "I/don't/exist"

        result = self.ha_service._change_cluster_participation(
            None, self.scripts_dir, self.ha_service.STATE_ONLINE)
        result.addCallbacks(check_success_result, should_not_be_called)

        # Now test the cluster standby script
        result = self.ha_service._change_cluster_participation(
            None, self.scripts_dir, self.ha_service.STATE_STANDBY)
        result.addCallbacks(check_success_result, should_not_be_called)
        return result

    def test_failed_cluster_standby_or_cluster_online_scripts(self):
        def expected_failure(result, script_path):
            self.assertEqual(
                str(result.value),
                "Failed charm script: %s exited with return code 2." %
                (script_path))

        def check_success_result(result):
            self.fail(
                "_change_cluster_participation ignored charm script failure.")

        # Rewrite both cluster scripts as failures
        for script_name in [
            self.ha_service.CLUSTER_ONLINE, self.ha_service.CLUSTER_STANDBY]:

            cluster_online = file(
                "%s/%s" % (self.scripts_dir, script_name), "w")
            cluster_online.write("#!/bin/bash\nexit 2")
            cluster_online.close()

        result = self.ha_service._change_cluster_participation(
            None, self.scripts_dir, self.ha_service.STATE_ONLINE)
        result.addCallback(check_success_result)
        script_path = (
            "%s/%s" % (self.scripts_dir, self.ha_service.CLUSTER_ONLINE))
        result.addErrback(expected_failure, script_path)

        # Now test the cluster standby script
        result = self.ha_service._change_cluster_participation(
            None, self.scripts_dir, self.ha_service.STATE_STANDBY)
        result.addCallback(check_success_result)
        script_path = (
            "%s/%s" % (self.scripts_dir, self.ha_service.CLUSTER_STANDBY))
        result.addErrback(expected_failure, script_path)
        return result

    def test_run_success_cluster_standby(self):
        """
        When receives a C{change-ha-service message} with C{STATE_STANDBY}
        requested the manager runs the C{CLUSTER_STANDBY} script and returns
        a successful operation-result to the server.
        """
        message = ({"type": "change-ha-service", "service-name": "my-service",
                    "unit-name": self.unit_name,
                    "service-state": self.ha_service.STATE_STANDBY,
                    "operation-id": 1})
        deferred = Deferred()

        def validate_messages(value):
            cluster_script = "%s/%s" % (
                self.scripts_dir, self.ha_service.CLUSTER_STANDBY)
            service = self.broker_service
            self.assertMessages(
                service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "result-text": u"%s succeeded." % cluster_script,
                  "status": SUCCEEDED, "operation-id": 1}])

        def handle_has_run(handle_result_deferred):
            handle_result_deferred.chainDeferred(deferred)
            return deferred.addCallback(validate_messages)

        ha_service_mock = self.mocker.patch(self.ha_service)
        ha_service_mock.handle_change_ha_service(ANY)
        self.mocker.passthrough(handle_has_run)
        self.mocker.replay()
        self.manager.add(self.ha_service)
        self.manager.dispatch_message(message)

        return deferred

    def test_run_success_cluster_online(self):
        """
        When receives a C{change-ha-service message} with C{STATE_ONLINE}
        requested the manager runs the C{CLUSTER_ONLINE} script and returns
        a successful operation-result to the server.
        """
        message = ({"type": "change-ha-service", "service-name": "my-service",
                    "unit-name": self.unit_name,
                    "service-state": self.ha_service.STATE_ONLINE,
                    "operation-id": 1})
        deferred = Deferred()

        def validate_messages(value):
            cluster_script = "%s/%s" % (
                self.scripts_dir, self.ha_service.CLUSTER_ONLINE)
            service = self.broker_service
            self.assertMessages(
                service.message_store.get_pending_messages(),
                [{"type": "operation-result",
                  "result-text": u"%s succeeded." % cluster_script,
                  "status": SUCCEEDED, "operation-id": 1}])

        def handle_has_run(handle_result_deferred):
            handle_result_deferred.chainDeferred(deferred)
            return deferred.addCallback(validate_messages)

        ha_service_mock = self.mocker.patch(self.ha_service)
        ha_service_mock.handle_change_ha_service(ANY)
        self.mocker.passthrough(handle_has_run)
        self.mocker.replay()
        self.manager.add(self.ha_service)
        self.manager.dispatch_message(message)

        return deferred



"""Helpers for reliable persistent message queues."""

ANCIENT = 1


def got_next_expected(store, next_expected):
    """Our peer has told us what it expects our next message's sequence to be.

    Call this with the message store and sequence number that the peer
    wants next; this will do various things based on what *this* side
    has in its outbound queue store.

    1. The peer expects a sequence greater than what we last
       sent. This is the common case and generally it should be
       expecting last_sent_sequence+len(messages_sent)+1.

    2. The peer expects a sequence number our side has already sent,
       and we no longer have that message. In this case, just send
       *all* messages we have, including the previous generation,
       starting at the sequence number the peer expects (meaning that
       messages have probably been lost).

    3. The peer expects a sequence number we already sent, and we
       still have that message cached. In this case, we send starting
       from that message.

    If the next expected sequence from the server refers to a message
    older than we have, then L{ANCIENT} will be returned.
    """
    ret = None
    old_sequence = store.get_sequence()
    if next_expected > old_sequence:
        store.delete_old_messages()
        pending_offset = next_expected - old_sequence
    elif next_expected < (old_sequence - store.get_pending_offset()):
        # "Ancient": The other side wants messages we don't have,
        # so let's just reset our counter to what it expects.
        pending_offset = 0
        ret = ANCIENT
    else:
        # No messages transferred, or
        # "Old": We'll try to send these old messages that the
        # other side still wants.
        pending_offset = (store.get_pending_offset() + next_expected
                          - old_sequence)

    store.set_pending_offset(pending_offset)
    store.set_sequence(next_expected)
    return ret



"""
This module contains a class, L{ConfigurationProxy} which pretends to be a
L{landscape.configuration.LandscapeSetupConfiguration} but actually directs
its calls via DBus to the L{ConfigurationMechanism}.
"""

import dbus

from landscape.ui.model.configuration.mechanism import (
    SERVICE_NAME, INTERFACE_NAME, OBJECT_PATH)
from landscape.configuration import LandscapeSetupConfiguration


class ConfigurationProxy(object):
    """
    L{ConfigurationProxy} attempts to be a drop-in replacement for
    L{LandscapeSetupConfiguration} allowing applications run by user accounts
    with the correct rights (as defined by a PolicyKit policy file) to interact
    with the landscape client configuration via a DBus service.  This is the
    RightThing(TM) for PolicyKit and therefore for GNOME/Unity.

    The canonical case for this is L{landscape-client-settings-ui}.
    """

    def __init__(self, bus=None, interface=None, loadargs=[]):
        self._loadargs = loadargs
        if bus is None:
            self._bus = dbus.SystemBus()
        else:
            self._bus = bus
        if interface is None:
            remote_object = self._bus.get_object(SERVICE_NAME, OBJECT_PATH)
            self._interface = dbus.Interface(remote_object, INTERFACE_NAME)
        else:
            self._interface = interface

    def load(self, arglist):
        if arglist is None:
            arglist = self._loadargs
        if len(arglist) == 0:
            al = ""
        else:
            al = chr(0x1e).join(arglist)
        try:
            self._interface.load(al)
        except dbus.DBusException, e:
            error_name = e.get_dbus_name()
            if error_name not in ("com.canonical.LandscapeClientSettings."
                                  "PermissionDeniedByPolicy",
                                  "org.freedesktop.DBus.Error.NoReply"):
                raise
            return False
        return True
    load.__doc__ = LandscapeSetupConfiguration.load.__doc__

    def reload(self):
        self._interface.reload()
    reload.__doc__ = LandscapeSetupConfiguration.reload.__doc__

    def write(self):
        self._interface.write()
    write.__doc__ = LandscapeSetupConfiguration.write.__doc__

    def get_config_filename(self):
        return self._interface.get_config_filename()
    get_config_filename.__doc__ = \
        LandscapeSetupConfiguration.get_config_filename.__doc__

    def exit(self, asynchronous=True):
        """
        Cause the mechanism to exit.
        """
        def on_reply():
            """
            This will never get called because we call L{sys.exit} inside the
            mechanism.
            """

        def on_error():
            """
            This will always be called, this allows us to supress the
            L{NoReply} error from DBus when we terminate the mechanism.
            """

        if asynchronous:
            self._interface.exit(reply_handler=on_reply,
                                 error_handler=on_error)
        else:
            self._interface.exit()

    def _delegate_to_interface(field):

        def get(self):
            return self._interface.get(field)

        def set(self, value):
            self._interface.set(field, value)

        return get, set

    account_name = property(*_delegate_to_interface("account_name"))
    computer_title = property(*_delegate_to_interface("computer_title"))
    data_path = property(*_delegate_to_interface("data_path"))
    http_proxy = property(*_delegate_to_interface("http_proxy"))
    https_proxy = property(*_delegate_to_interface("https_proxy"))
    ping_url = property(*_delegate_to_interface("ping_url"))
    registration_key = property(
        *_delegate_to_interface("registration_key"))
    tags = property(*_delegate_to_interface("tags"))
    url = property(*_delegate_to_interface("url"))



from twisted.internet.defer import succeed

from landscape.monitor.plugin import MonitorPlugin, DataWatcher
from landscape.schema import Message, Int
from landscape.tests.mocker import ANY
from landscape.tests.helpers import (
    LandscapeTest, MonitorHelper, LogKeeperHelper)


class MonitorPluginTest(LandscapeTest):

    helpers = [MonitorHelper]

    def test_without_persist_name(self):
        """
        By default a L{MonitorPlugin} doesn't have a C{_persist} attribute.
        """
        plugin = MonitorPlugin()
        plugin.register(self.monitor)
        self.assertIs(plugin.persist, None)

    def test_with_persist_name(self):
        """
        When plugins providea C{persist_name} attribute, they get a persist
        object set at C{_persist} which is rooted at the name specified.
        """
        plugin = MonitorPlugin()
        plugin.persist_name = "wubble"
        plugin.register(self.monitor)
        plugin.persist.set("hi", "there")
        self.assertEqual(self.monitor.persist.get("wubble"), {"hi": "there"})

    def test_with_no_run_interval(self):
        """
        If the C{run_interval} attribute of L{MonitorPlugin} is C{None}, its
        C{run} method won't get called by the reactor.
        """
        plugin = MonitorPlugin()
        plugin.run = lambda: 1 / 0
        plugin.run_interval = None
        plugin.register(self.monitor)
        self.reactor.advance(MonitorPlugin.run_interval)

    def test_call_on_accepted(self):
        """
        L{MonitorPlugin}-based plugins can provide a callable to call
        when a message type becomes accepted.
        """
        plugin = MonitorPlugin()
        plugin.register(self.monitor)
        callback = self.mocker.mock()
        callback("foo", kwarg="bar")
        self.mocker.replay()
        plugin.call_on_accepted("type", callback, "foo", kwarg="bar")
        self.reactor.fire(("message-type-acceptance-changed", "type"), True)

    def test_call_on_accepted_when_unaccepted(self):
        """
        Notifications are only dispatched to plugins when types become
        accepted, not when they become unaccepted.
        """
        plugin = MonitorPlugin()
        plugin.register(self.monitor)
        callback = lambda: 1 / 0
        plugin.call_on_accepted("type", callback)
        self.reactor.fire(("message-type-acceptance-changed", "type"), False)

    def test_resynchronize_with_global_scope(self):
        """
        If a 'resynchronize' event fires with global scope, we clear down the
        persist.
        """
        plugin = MonitorPlugin()
        plugin.persist_name = "wubble"
        plugin.register(self.monitor)
        plugin.persist.set("hi", "there")
        self.assertEqual(self.monitor.persist.get("wubble"), {"hi": "there"})
        self.reactor.fire("resynchronize")
        self.assertTrue(self.monitor.persist.get("wubble") is None)

    def test_resynchronize_with_provided_scope(self):
        """
        If a 'resynchronize' event fires with the provided scope, we clear down
        the persist.
        """
        plugin = MonitorPlugin()
        plugin.persist_name = "wubble"
        plugin.scope = "frujical"
        plugin.register(self.monitor)
        plugin.persist.set("hi", "there")
        self.assertEqual(self.monitor.persist.get("wubble"), {"hi": "there"})
        self.reactor.fire("resynchronize", scopes=["frujical"])
        self.assertTrue(self.monitor.persist.get("wubble") is None)

    def test_do_not_resynchronize_with_other_scope(self):
        """
        If a 'resynchronize' event fires with an irrelevant scope, we do
        nothing.
        """
        plugin = MonitorPlugin()
        plugin.persist_name = "wubble"
        plugin.scope = "frujical"
        plugin.register(self.monitor)
        plugin.persist.set("hi", "there")
        self.assertEqual(self.monitor.persist.get("wubble"), {"hi": "there"})
        self.reactor.fire("resynchronize", scopes=["chrutfup"])
        self.assertEqual(self.monitor.persist.get("wubble"), {"hi": "there"})


class StubDataWatchingPlugin(DataWatcher):

    persist_name = "ooga"
    message_type = "wubble"
    message_key = "wubblestuff"

    def __init__(self, data=None):
        self.data = data

    def get_data(self):
        return self.data


class DataWatcherTest(LandscapeTest):

    helpers = [MonitorHelper, LogKeeperHelper]

    def setUp(self):
        LandscapeTest.setUp(self)
        self.plugin = StubDataWatchingPlugin(1)
        self.plugin.register(self.monitor)
        self.mstore.add_schema(Message("wubble", {"wubblestuff": Int()}))

    def test_get_message(self):
        self.assertEqual(self.plugin.get_message(),
                         {"type": "wubble", "wubblestuff": 1})

    def test_get_message_unchanging(self):
        self.assertEqual(self.plugin.get_message(),
                         {"type": "wubble", "wubblestuff": 1})
        self.assertEqual(self.plugin.get_message(), None)

    def test_basic_exchange(self):
        # Is this really want we want to do?
        self.mstore.set_accepted_types(["wubble"])
        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(messages[0]["type"], "wubble")
        self.assertEqual(messages[0]["wubblestuff"], 1)
        self.assertIn("Queueing a message with updated data watcher info for "
                      "landscape.monitor.tests.test_plugin.StubDataWatching"
                      "Plugin.", self.logfile.getvalue())

    def test_unchanging_value(self):
        # Is this really want we want to do?
        self.mstore.set_accepted_types(["wubble"])
        self.plugin.exchange()
        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)

    def test_urgent_exchange(self):
        """
        When exchange is called with an urgent argument set to True
        make sure it sends the message urgently.
        """
        remote_broker_mock = self.mocker.replace(self.remote)
        remote_broker_mock.send_message(ANY, ANY, urgent=True)
        self.mocker.result(succeed(None))
        self.mocker.replay()

        self.mstore.set_accepted_types(["wubble"])
        self.plugin.exchange(True)

    def test_no_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently
        accepting their type.
        """
        self.mstore.set_accepted_types([])
        self.reactor.advance(self.monitor.step_size * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["wubble"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])



import json
import os

from landscape.monitor.jujuinfo import JujuInfo
from landscape.tests.helpers import LandscapeTest, MonitorHelper


SAMPLE_JUJU_INFO = json.dumps({"environment-uuid": "DEAD-BEEF",
                               "unit-name": "juju-unit-name",
                               "api-addresses": "10.0.3.1:17070",
                               "private-address": "127.0.0.1"})


class JujuInfoTest(LandscapeTest):

    helpers = [MonitorHelper]

    def setUp(self):
        super(JujuInfoTest, self).setUp()
        self.mstore.set_accepted_types(["juju-info"])
        self.plugin = JujuInfo()
        self.monitor.add(self.plugin)
        self.makeFile(SAMPLE_JUJU_INFO, path=self.config.juju_filename)

    def test_get_sample_juju_info(self):
        """
        Sample data is used to ensure that expected values end up in
        the Juju data reported by the plugin.
        """
        self.plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "juju-info")
        self.assertEqual(message["environment-uuid"], "DEAD-BEEF")
        self.assertEqual(message["unit-name"], "juju-unit-name")
        self.assertEqual(message["api-addresses"], ["10.0.3.1:17070"])

    def test_juju_info_reported_only_once(self):
        """
        Juju data shouldn't be reported unless it's changed since the
        last time it was reported.
        """
        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["type"], "juju-info")

        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)

    def test_report_changed_juju_info(self):
        """
        When juju data changes, the new data should be sent to the
        server.
        """
        self.plugin.exchange()
        message = self.mstore.get_pending_messages()[0]
        self.assertEqual(message["type"], "juju-info")
        self.assertEqual(message["environment-uuid"], "DEAD-BEEF")
        self.assertEqual(message["unit-name"], "juju-unit-name")
        self.assertEqual(message["api-addresses"], ["10.0.3.1:17070"])

        self.makeFile(
            json.dumps({"environment-uuid": "FEED-BEEF",
                        "unit-name": "changed-unit-name",
                        "api-addresses": "10.0.3.2:17070",
                        "private-address": "127.0.1.1"}),
            path=self.config.juju_filename)
        self.plugin.exchange()
        message = self.mstore.get_pending_messages()[1]
        self.assertEqual(message["type"], "juju-info")
        self.assertEqual(message["environment-uuid"], "FEED-BEEF")
        self.assertEqual(message["unit-name"], "changed-unit-name")
        self.assertEqual(message["api-addresses"], ["10.0.3.2:17070"])

    def test_no_message_with_invalid_json(self):
        """No Juju message is sent if the JSON file is invalid."""
        self.makeFile("barf", path=self.config.juju_filename)

        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(messages, [])
        self.log_helper.ignore_errors(ValueError)
        self.assertTrue(
            "Error attempting to read JSON from" in self.logfile.getvalue())

    def test_no_message_with_missing_file(self):
        """No Juju message is sent if the JSON file is missing."""
        os.remove(self.config.juju_filename)

        self.plugin.exchange()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(messages, [])



from logging import info

from landscape.log import format_object


class PluginConfigError(Exception):
    """There was an error registering or configuring a plugin."""


class PluginRegistry(object):
    """A central integration point for plugins."""

    def __init__(self):
        self._plugins = []
        self._plugin_names = {}

    def add(self, plugin):
        """Register a plugin.

        The plugin's C{register} method will be called with this registry as
        its argument.

        If the plugin has a C{plugin_name} attribute, it will be possible to
        look up the plugin later with L{get_plugin}.
        """
        info("Registering plugin %s.", format_object(plugin))
        self._plugins.append(plugin)
        if hasattr(plugin, "plugin_name"):
            self._plugin_names[plugin.plugin_name] = plugin
        plugin.register(self)

    def get_plugins(self):
        """Get the list of plugins."""
        return self._plugins

    def get_plugin(self, name):
        """Get a particular plugin by name."""
        return self._plugin_names[name]


class Plugin(object):
    """A convenience for writing plugins.

    This provides a register method which will set up a bunch of
    reactor handlers in the idiomatic way.

    If C{run} is defined on subclasses, it will be called every C{run_interval}
    seconds after being registered.

    @cvar run_interval: The interval, in seconds, to execute the
    C{run} method. If set to C{None}, then C{run} will not be
    scheduled.
    """

    run_interval = 5

    def register(self, registry):
        self.registry = registry
        if hasattr(self, "run") and self.run_interval is not None:
            registry.reactor.call_every(self.run_interval, self.run)



from landscape.lib.amp import MethodCallError
from landscape.tests.helpers import LandscapeTest, DEFAULT_ACCEPTED_TYPES
from landscape.broker.tests.helpers import (
    RemoteBrokerHelper, RemoteClientHelper)


class RemoteBrokerTest(LandscapeTest):

    helpers = [RemoteBrokerHelper]

    def test_ping(self):
        """
        The L{RemoteBroker.ping} method calls the C{ping} method of the
        remote L{BrokerServer} instance and returns its result with a
        L{Deferred}.
        """
        result = self.remote.ping()
        return self.assertSuccess(result, True)

    def test_register_client(self):
        """
        The L{RemoteBroker.register_client} method forwards a registration
        request to the remote L{BrokerServer} object.
        """
        self.broker.register_client = self.mocker.mock()
        self.expect(self.broker.register_client("client"))
        self.mocker.replay()
        result = self.remote.register_client("client")
        return self.assertSuccess(result)

    def test_send_message(self):
        """
        The L{RemoteBroker.send_message} method calls the C{send_message}
        method of the remote L{BrokerServer} instance and returns its result
        with a L{Deferred}.
        """
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])

        session_id = self.successResultOf(self.remote.get_session_id())
        message_id = self.successResultOf(
            self.remote.send_message(message, session_id))

        self.assertTrue(isinstance(message_id, int))
        self.assertTrue(self.mstore.is_pending(message_id))
        self.assertFalse(self.exchanger.is_urgent())
        self.assertMessages(self.mstore.get_pending_messages(),
                            [message])

    def test_send_message_with_urgent(self):
        """
        The L{RemoteBroker.send_message} method honors the urgent argument.
        """
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])
        session_id = self.successResultOf(self.remote.get_session_id())
        message_id = self.successResultOf(self.remote.send_message(
            message, session_id, urgent=True))
        self.assertTrue(isinstance(message_id, int))
        self.assertTrue(self.exchanger.is_urgent())

    def test_is_message_pending(self):
        """
        The L{RemoteBroker.is_message_pending} method calls the
        C{is_message_pending} method of the remote L{BrokerServer} instance
        and returns its result with a L{Deferred}.
        """
        result = self.remote.is_message_pending(1234)
        return self.assertSuccess(result, False)

    def test_stop_clients(self):
        """
        The L{RemoteBroker.stop_clients} method calls the C{stop_clients}
        method of the remote L{BrokerServer} instance and returns its result
        with a L{Deferred}.
        """
        result = self.remote.stop_clients()
        return self.assertSuccess(result, None)

    def test_reload_configuration(self):
        """
        The L{RemoteBroker.reload_configuration} method calls the
        C{reload_configuration} method of the remote L{BrokerServer}
        instance and returns its result with a L{Deferred}.
        """
        result = self.remote.reload_configuration()
        return self.assertSuccess(result, None)

    def test_register(self):
        """
        The L{RemoteBroker.register} method calls the C{register} method
        of the remote L{BrokerServer} instance and returns its result with
        a L{Deferred}.
        """
        # This should make the registration succeed
        self.transport.responses.append([{"type": "set-id", "id": "abc",
                                          "insecure-id": "def"}])
        result = self.remote.register()
        return self.assertSuccess(result, None)

    def test_get_accepted_message_types(self):
        """
        The L{RemoteBroker.get_accepted_message_types} method calls the
        C{get_accepted_message_types} method of the remote L{BrokerServer}
        instance and returns its result with a L{Deferred}.
        """
        result = self.remote.get_accepted_message_types()
        return self.assertSuccess(result, self.mstore.get_accepted_types())

    def test_get_server_uuid(self):
        """
        The L{RemoteBroker.get_server_uuid} method calls the C{get_server_uuid}
        method of the remote L{BrokerServer} instance and returns its result
        with a L{Deferred}.
        """
        self.mstore.set_server_uuid("abcde")
        result = self.remote.get_server_uuid()
        return self.assertSuccess(result, "abcde")

    def test_register_client_accepted_message_type(self):
        """
        The L{RemoteBroker.register_client_accepted_message_type} method calls
        the C{register_client_accepted_message_type} method of the remote
        L{BrokerServer} instance and returns its result with a L{Deferred}.
        """

        def assert_response(response):
            self.assertEqual(response, None)
            self.assertEqual(
                self.exchanger.get_client_accepted_message_types(),
                sorted(["type"] + DEFAULT_ACCEPTED_TYPES))

        result = self.remote.register_client_accepted_message_type("type")
        return result.addCallback(assert_response)

    def test_exit(self):
        """
        The L{RemoteBroker.exit} method calls the C{exit} method of the remote
        L{BrokerServer} instance and returns its result with a L{Deferred}.
        """
        result = self.remote.exit()
        return self.assertSuccess(result, None)

    def test_call_if_accepted(self):
        """
        The L{RemoteBroker.call_if_accepted} method calls a function if the
        given message type is accepted.
        """
        self.mstore.set_accepted_types(["test"])
        function = self.mocker.mock()
        self.expect(function(123)).result("cool")
        self.mocker.replay()
        result = self.remote.call_if_accepted("test", function, 123)
        return self.assertSuccess(result, "cool")

    def test_call_if_accepted_with_not_accepted(self):
        """
        The L{RemoteBroker.call_if_accepted} method doesn't do anything if the
        given message type is not accepted.
        """
        function = lambda: 1 / 0
        result = self.remote.call_if_accepted("test", function)
        return self.assertSuccess(result, None)

    def test_listen_events(self):
        """
        L{RemoteBroker.listen_events} returns a deferred which fires when
        the first of the given events occurs in the broker reactor.
        """
        deferred = self.remote.listen_events(["event1", "event2"])
        self.reactor.call_later(0.05, self.reactor.fire, "event2")
        self.reactor.advance(0.05)
        self.remote._factory.fake_connection.flush()
        self.assertEqual("event2", self.successResultOf(deferred))

    def test_call_on_events(self):
        """
        L{RemoteBroker.call_on_events} fires the given callback when the
        first of the given events occurs in the broker reactor.
        """
        callback1 = self.mocker.mock()
        self.expect(callback1()).count(0)
        callback2 = self.mocker.mock()
        self.expect(callback2()).result(123)
        self.mocker.replay()
        deferred = self.remote.call_on_event({"event1": callback1,
                                              "event2": callback2})
        self.reactor.call_later(0.05, self.reactor.fire, "event2")
        self.reactor.advance(0.05)
        self.remote._factory.fake_connection.flush()
        self.assertEqual(123, self.successResultOf(deferred))

    def test_fire_event(self):
        """
        The L{RemoteBroker.fire_event} method fires an event in the broker
        reactor.
        """
        callback = self.mocker.mock()
        callback()
        self.mocker.replay()
        self.reactor.call_on("event", callback)
        return self.assertSuccess(self.remote.fire_event("event"))

    def test_method_call_error(self):
        """
        Trying to call an non-exposed broker method results in a failure.
        """
        deferred = self.remote.get_clients()
        self.failureResultOf(deferred).trap(MethodCallError)


class RemoteClientTest(LandscapeTest):

    helpers = [RemoteClientHelper]

    def test_ping(self):
        """
        The L{RemoteClient.ping} method calls the C{ping} method of the
        remote L{BrokerClient} instance and returns its result with a
        L{Deferred}.
        """
        result = self.remote_client.ping()
        return self.assertSuccess(result, True)

    def test_message(self):
        """
        The L{RemoteClient.message} method calls the C{message} method of
        the remote L{BrokerClient} instance and returns its result with
        a L{Deferred}.
        """
        handler = self.mocker.mock()
        handler({"type": "test"})
        self.client.broker = self.mocker.mock()
        self.client.broker.register_client_accepted_message_type("test")
        self.mocker.replay()

        # We need to register a test message handler to let the dispatch
        # message method call succeed
        self.client.register_message("test", handler)
        result = self.remote_client.message({"type": "test"})
        return self.assertSuccess(result, True)

    def test_fire_event(self):
        """
        The L{RemoteClient.fire_event} method calls the C{fire_event} method of
        the remote L{BrokerClient} instance and returns its result with a
        L{Deferred}.
        """
        callback = self.mocker.mock()
        callback(True, kwarg=2)
        self.mocker.replay()
        self.client_reactor.call_on("event", callback)
        result = self.remote_client.fire_event("event", True, kwarg=2)
        return self.assertSuccess(result, [None])

    def test_exit(self):
        """
        The L{RemoteClient.exit} method calls the C{exit} method of the remote
        L{BrokerClient} instance and returns its result with a L{Deferred}.
        """
        result = self.remote_client.exit()
        return self.assertSuccess(result, None)

    def test_method_call_error(self):
        """
        Trying to call an non-exposed client method results in a failure.
        """
        deferred = self.remote_client.get_plugins()
        self.failureResultOf(deferred).trap(MethodCallError)



from landscape.monitor.updatemanager import UpdateManager
from landscape.tests.helpers import (
    LandscapeTest, MonitorHelper, LogKeeperHelper)
from landscape.tests.mocker import ANY


class UpdateManagerTest(LandscapeTest):
    """
    Tests relating to the L{UpdateManager} monitoring plug-in, which should
    notice changes to update-manager's configuration and report these back to
    landscape server.
    """

    helpers = [MonitorHelper, LogKeeperHelper]

    def setUp(self):
        super(UpdateManagerTest, self).setUp()
        self.update_manager_filename = self.makeFile()
        self.plugin = UpdateManager(self.update_manager_filename)
        self.monitor.add(self.plugin)
        self.mstore.set_accepted_types(["update-manager-info"])

    def test_get_prompt(self):
        """
        L{UpdateManager._get_prompt} returns the value of the
        variable C{Prompt} in the update-manager's configuration.
        """
        content = """
[DEFAULT]
Prompt=lts
"""
        self.makeFile(path=self.update_manager_filename, content=content)
        self.assertEqual("lts", self.plugin._get_prompt())

    def test_get_prompt_with_invalid_value_configured(self):
        """
        L{update_manager._get_prompt} returns "normal" if an invalid value
        is specified in the file.  A warning is also logged.
        """
        content = """
[DEFAULT]
Prompt=zarniwhoop
"""
        self.makeFile(path=self.update_manager_filename, content=content)
        self.assertEqual("normal", self.plugin._get_prompt())

    def test_get_prompt_with_missing_config_file(self):
        """
        When the configuration file does not exist we just return "normal".
        Any machine that doesn't have update-manager installed would fit into
        this category, so there's no need to warn about it.
        """
        self.plugin.update_manager_filename = "/I/Do/Not/Exist"
        self.assertEqual("normal", self.plugin._get_prompt())

    def test_send_message(self):
        """
        A new C{"update-manager-info"} message should be enqueued if and only
        if the update-manager status of the system has changed.
        """
        content = """
[DEFAULT]
Prompt=never
"""
        self.makeFile(path=self.update_manager_filename, content=content)
        self.plugin.send_message()
        self.assertIn("Queueing message with updated update-manager status.",
                      self.logfile.getvalue())
        self.assertMessages(self.mstore.get_pending_messages(),
                            [{"type": "update-manager-info",
                              "prompt": u"never"}])
        self.mstore.delete_all_messages()
        self.plugin.send_message()
        self.assertMessages(self.mstore.get_pending_messages(), [])

    def test_run_interval(self):
        """
        The L{UpdateManager} plugin will be scheduled to run every hour.
        """
        self.assertEqual(3600, self.plugin.run_interval)

    def test_run_immediately(self):
        """
        The L{UpdateManager} plugin will be run immediately at startup.
        """
        self.assertTrue(True, self.plugin.run_immediately)

    def test_run(self):
        """
        If the server can accept them, the plugin should send
        C{update-manager} messages.
        """
        broker_mock = self.mocker.replace(self.remote)
        broker_mock.send_message(ANY, ANY)
        self.mocker.replay()
        self.plugin.run()
        self.mstore.set_accepted_types([])
        self.plugin.run()

    def test_resynchronize(self):
        """
        The "resynchronize" reactor message cause the plugin to send fresh
        data.
        """
        self.plugin.run()
        self.reactor.fire("resynchronize", scopes=["package"])
        self.plugin.run()
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 2)



import copy

from landscape.lib.network import get_fqdn

from landscape.ui.constants import CANONICAL_MANAGED, NOT_MANAGED
from landscape.ui.model.configuration.proxy import ConfigurationProxy


HOSTED_LANDSCAPE_HOST = "landscape.canonical.com"
LOCAL_LANDSCAPE_HOST = ""

HOSTED_ACCOUNT_NAME = ""
LOCAL_ACCOUNT_NAME = "standalone"

HOSTED_PASSWORD = ""
LOCAL_PASSWORD = ""

HOSTED = "hosted"
LOCAL = "local"
MANAGEMENT_TYPE = "management-type"
COMPUTER_TITLE = "computer-title"
LANDSCAPE_HOST = "landscape-host"
ACCOUNT_NAME = "account-name"
PASSWORD = "password"


DEFAULT_DATA = {
    MANAGEMENT_TYPE: NOT_MANAGED,
    COMPUTER_TITLE: get_fqdn(),
    HOSTED: {
        LANDSCAPE_HOST: HOSTED_LANDSCAPE_HOST,
        ACCOUNT_NAME: HOSTED_ACCOUNT_NAME,
        PASSWORD: HOSTED_PASSWORD},
    LOCAL: {
        LANDSCAPE_HOST: LOCAL_LANDSCAPE_HOST,
        ACCOUNT_NAME: LOCAL_ACCOUNT_NAME,
        PASSWORD: LOCAL_PASSWORD}}


def derive_server_host_name_from_url(url):
    """
    Extract the hostname part from a URL.
    """
    try:
        without_protocol = url[url.index("://") + 3:]
    except ValueError:
        without_protocol = url
    try:
        return without_protocol[:without_protocol.index("/")]
    except ValueError:
        return without_protocol


def derive_url_from_host_name(host_name):
    """
    Extrapolate a url from a host name.
    """
    #Reuse this code to make sure it's a proper host name
    host_name = derive_server_host_name_from_url(host_name)
    return "https://" + host_name + "/message-system"


def derive_ping_url_from_host_name(host_name):
    """
    Extrapolate a ping_url from a host name.
    """
    #Reuse this code to make sure it's a proper host name
    host_name = derive_server_host_name_from_url(host_name)
    return "http://" + host_name + "/ping"


class StateError(Exception):
    """
    An exception that is raised when there is an error relating to the current
    state.
    """


class TransitionError(Exception):
    """
    An L{Exception} that is raised when a valid transition between states fails
    for some non state related reason.  For example, this error is raised when
    the user does not have the privilege of reading the configuration file,
    this causes the transition from L{VirginState} to L{InitialisedState} to
    fail but not because that transition from one state to another was not
    permitted, but rather the transition encountered an error.
    """


class ConfigurationState(object):
    """
    Base class for states used in the L{ConfigurationModel}.
    """
    def __init__(self, data, proxy, uisettings):
        self._data = copy.deepcopy(data)
        self._proxy = proxy
        self._uisettings = uisettings

    def get_config_filename(self):
        return self._proxy.get_config_filename()

    def get(self, *args):
        """
        Retrieve only valid values from two level dictionary based tree.

        This mainly served to pick up programming errors and could easily be
        replaced with a simpler scheme.
        """
        arglen = len(args)
        if arglen > 2 or arglen == 0:
            raise TypeError(
                "get() takes either 1 or 2 keys (%d given)" % arglen)
        if arglen == 2:  # We're looking for a leaf on a branch
            sub_dict = None
            if args[0] in [HOSTED, LOCAL]:
                sub_dict = self._data.get(args[0], {})
            sub_dict = self._data[args[0]]
            if not isinstance(sub_dict, dict):
                raise KeyError(
                    "Compound key [%s][%s] is invalid. The data type " +
                    "returned from the first index was %s." %
                    sub_dict.__class__.__name__)
            return sub_dict.get(args[1], None)
        else:
            if args[0] in (MANAGEMENT_TYPE, COMPUTER_TITLE):
                return self._data.get(args[0], None)
            else:
                raise KeyError("Key [%s] is invalid. " % args[0])

    def set(self, *args):
        """
        Set only valid values from two level dictionary based tree.

        This mainly served to pick up programming errors and could easily be
        replaced with a simpler scheme.
        """
        arglen = len(args)
        if arglen < 2 or arglen > 3:
            raise TypeError("set() takes either 1 or 2 keys and exactly 1 " +
                            "value (%d arguments given)" % arglen)
        if arglen == 2:  # We're setting a leaf attached to the root
            self._data[args[0]] = args[1]
        else:  # We're setting a leaf on a branch
            sub_dict = None
            if args[0] in [HOSTED, LOCAL]:
                sub_dict = self._data.get(args[0], {})
            if not isinstance(sub_dict, dict):
                raise KeyError("Compound key [%s][%s] is invalid. The data " +
                               "type returned from the first index was %s."
                               % sub_dict.__class__.__name__)
            sub_dict[args[1]] = args[2]
            self._data[args[0]] = sub_dict

    def load_data(self, asynchronous=True, exit_method=None):
        raise NotImplementedError

    def modify(self):
        raise NotImplementedError

    def revert(self):
        raise NotImplementedError

    def persist(self):
        raise NotImplementedError

    def exit(self, asynchronous=True, exit_method=None):
        return ExitedState(self._data, self._proxy, self._uisettings,
                           asynchronous=asynchronous, exit_method=exit_method)


class Helper(object):
    """
    Base class for all state transition helpers.

    It is assumed that the Helper classes are "friends" of the
    L{ConfigurationState} classes and can have some knowledge of their
    internals.  They shouldn't be visible to users of the
    L{ConfigurationState}s and in general we should avoid seeing the
    L{ConfigurationState}'s _data attribute outside this module.
    """

    def __init__(self, state):
        self._state = state


class ModifiableHelper(Helper):
    """
    Allow a L{ConfigurationState}s to be modified.
    """

    def modify(self):
        return ModifiedState(self._state._data, self._state._proxy,
                             self._state._uisettings)


class UnloadableHelper(Helper):
    """
    Disallow loading of data into a L{ConfigurationModel}.
    """

    def load_data(self, asynchronous=True, exit_method=None):
        raise StateError("A ConfiguratiomModel in a " +
                         self.__class__.__name__ +
                         " cannot be transitioned via load_data()")


class UnmodifiableHelper(Helper):
    """
    Disallow modification of a L{ConfigurationState}.
    """

    def modify(self):
        raise StateError("A ConfigurationModel in " +
                         self.__class__.__name__ +
                         " cannot transition via modify()")


class RevertableHelper(Helper):
    """
    Allow reverting of a L{ConfigurationModel}.
    """

    def revert(self):
        return InitialisedState(self._state._data, self._state._proxy,
                                self._state._uisettings)


class UnrevertableHelper(Helper):
    """
    Disallow reverting of a L{ConfigurationModel}.
    """

    def revert(self):
        raise StateError("A ConfigurationModel in " +
                         self.__class__.__name__ +
                         " cannot transition via revert()")


class PersistableHelper(Helper):
    """
    Allow a L{ConfigurationModel} to persist.
    """

    def _save_to_uisettings(self):
        """
        Persist full content to the L{UISettings} object.
        """
        self._state._uisettings.set_management_type(
            self._state.get(MANAGEMENT_TYPE))
        self._state._uisettings.set_computer_title(
            self._state.get(COMPUTER_TITLE))
        self._state._uisettings.set_hosted_account_name(
            self._state.get(HOSTED, ACCOUNT_NAME))
        self._state._uisettings.set_hosted_password(
            self._state.get(HOSTED, PASSWORD))
        self._state._uisettings.set_local_landscape_host(
            self._state.get(LOCAL, LANDSCAPE_HOST))
        self._state._uisettings.set_local_account_name(
            self._state.get(LOCAL, ACCOUNT_NAME))
        self._state._uisettings.set_local_password(
            self._state.get(LOCAL, PASSWORD))

    def _save_to_config(self):
        """
        Persist the subset of the data we want to make live to the actual
        configuration file.
        """
        hosted = self._state.get(MANAGEMENT_TYPE)
        if hosted is NOT_MANAGED:
            pass
        else:
            if hosted == CANONICAL_MANAGED:
                first_key = HOSTED
            else:
                first_key = LOCAL
            self._state._proxy.url = derive_url_from_host_name(
                self._state.get(first_key, LANDSCAPE_HOST))
            self._state._proxy.ping_url = derive_ping_url_from_host_name(
                self._state.get(first_key, LANDSCAPE_HOST))
            self._state._proxy.account_name = self._state.get(
                first_key, ACCOUNT_NAME)
            self._state._proxy.registration_key = self._state.get(
                first_key, PASSWORD)
            self._state._proxy.computer_title = self._state.get(COMPUTER_TITLE)
            self._state._proxy.write()

    def persist(self):
        self._save_to_uisettings()
        self._save_to_config()
        return InitialisedState(self._state._data, self._state._proxy,
                                self._state._uisettings)


class UnpersistableHelper(Helper):
    """
    Disallow persistence of a L{ConfigurationModel}.
    """

    def persist(self):
        raise StateError("A ConfiguratonModel in " +
                         self.__class__.__name__ +
                         " cannot be transitioned via persist().")


class ExitedState(ConfigurationState):
    """
    The terminal state of L{ConfigurationModel}, you can't do anything further
    once this state is reached.
    """
    def __init__(self, data, proxy, uisettings, exit_method=None,
                 asynchronous=True):
        super(ExitedState, self).__init__(None, None, None)
        if callable(exit_method):
            exit_method()
        else:
            proxy.exit(asynchronous=asynchronous)
        self._unloadable_helper = UnloadableHelper(self)
        self._unmodifiable_helper = UnmodifiableHelper(self)
        self._unrevertable_helper = UnrevertableHelper(self)
        self._unpersistable_helper = UnpersistableHelper(self)

    def load_data(self, asynchronous=True, exit_method=None):
        return self._unloadable_helper.load_data(asynchronous=asynchronous,
                                                 exit_method=exit_method)

    def modify(self):
        return self._unmodifiable_helper.modify()

    def revert(self):
        return self._unrevertable_helper.revert()

    def persist(self):
        return self._unpersistable_helper.persist()

    def exit(self, asynchronous=True):
        return self


class ModifiedState(ConfigurationState):
    """
    The state of a L{ConfigurationModel} whenever the user has modified some
    data but hasn't yet L{persist}ed or L{revert}ed.
    """

    def __init__(self, data, proxy, uisettings):
        super(ModifiedState, self).__init__(data, proxy, uisettings)
        self._modifiable_helper = ModifiableHelper(self)
        self._revertable_helper = RevertableHelper(self)
        self._persistable_helper = PersistableHelper(self)

    def modify(self):
        return self._modifiable_helper.modify()

    def revert(self):
        return self._revertable_helper.revert()

    def persist(self):
        return self._persistable_helper.persist()


class InitialisedState(ConfigurationState):
    """
    The state of the L{ConfigurationModel} as initially presented to the
    user. Baseline data should have been loaded from the real configuration
    data, any persisted user data should be loaded into blank values and
    finally defaults should be applied where necessary.
    """

    def __init__(self, data, proxy, uisettings):
        super(InitialisedState, self).__init__(data, proxy, uisettings)
        self._modifiable_helper = ModifiableHelper(self)
        self._unrevertable_helper = UnrevertableHelper(self)
        self._unpersistable_helper = UnpersistableHelper(self)
        self._load_uisettings_data()
        if not self._load_live_data():
            raise TransitionError("Authentication Failure")

    def _load_uisettings_data(self):
        """
        Load the complete set of dialog data from L{UISettings}.
        """
        hosted = self._uisettings.get_management_type()
        self.set(MANAGEMENT_TYPE, hosted)
        computer_title = self._uisettings.get_computer_title()
        if computer_title:
            self.set(COMPUTER_TITLE, computer_title)
        self.set(HOSTED, ACCOUNT_NAME,
                 self._uisettings.get_hosted_account_name())
        self.set(HOSTED, PASSWORD, self._uisettings.get_hosted_password())
        self.set(LOCAL, LANDSCAPE_HOST,
                 self._uisettings.get_local_landscape_host())
        local_account_name = self._uisettings.get_local_account_name()
        if local_account_name:
            self.set(LOCAL, ACCOUNT_NAME, local_account_name)
        self.set(LOCAL, PASSWORD, self._uisettings.get_local_password())

    def _load_live_data(self):
        """
        Load the current live subset of data from the configuration file.
        """
        if self._proxy.load(None):
            computer_title = self._proxy.computer_title
            if computer_title:
                self.set(COMPUTER_TITLE, computer_title)
            url = self._proxy.url
            if url.find(HOSTED_LANDSCAPE_HOST) > -1:
                self.set(HOSTED, ACCOUNT_NAME, self._proxy.account_name)
                self.set(HOSTED, PASSWORD, self._proxy.registration_key)
            else:
                self.set(LOCAL, LANDSCAPE_HOST,
                         derive_server_host_name_from_url(url))
                if self._proxy.account_name != "":
                    self.set(LOCAL, ACCOUNT_NAME, self._proxy.account_name)
            return True
        else:
            return False

    def load_data(self, asynchronous=True, exit_method=None):
        return self

    def modify(self):
        return self._modifiable_helper.modify()

    def revert(self):
        return self._unrevertable_helper.revert()

    def persist(self):
        return self._unpersistable_helper.persist()


class VirginState(ConfigurationState):
    """
    The state of the L{ConfigurationModel} before any actions have been taken
    upon it.
    """

    def __init__(self, proxy, uisettings):
        super(VirginState, self).__init__(DEFAULT_DATA, proxy, uisettings)
        self._unmodifiable_helper = UnmodifiableHelper(self)
        self._unrevertable_helper = UnrevertableHelper(self)
        self._unpersistable_helper = UnpersistableHelper(self)

    def load_data(self, asynchronous=True, exit_method=None):
        try:
            return InitialisedState(self._data, self._proxy, self._uisettings)
        except TransitionError:
            return ExitedState(self._data, self._proxy, self._uisettings,
                               asynchronous=asynchronous,
                               exit_method=exit_method)

    def modify(self):
        return self._unmodifiable_helper.modify()

    def revert(self):
        return self._unrevertable_helper.revert()

    def persist(self):
        return self._unpersistable_helper.persist()


class ConfigurationModel(object):
    """
    L{ConfigurationModel} presents a model of configuration as the UI
    requirements describe it (separate values for the Hosted and Local
    configurations) as opposed to the real structure of the configuration
    file.  This is intended to achieve the following:

       1. Allow the expected behaviour in the UI without changing the live
          config file.
       2. Supersede the overly complex logic in the controller layer with a
          cleaner state pattern.

    The allowable state transitions are:

       VirginState      --(load_data)--> InitialisedState
       VirginState      --(load_data)--> ExitedState
       VirginState      --(exit)-------> ExitedState
       InitialisedState --(modify)-----> ModifiedState
       InitialisedState --(exit)-------> ExitedState
       ModifiedState    --(revert)-----> InitialisedState
       ModifiedState    --(modify)-----> ModifiedState
       ModifiedState    --(persist)----> InitialisedState
       ModifiedState    --(exit)-------> ExitedState
    """

    def __init__(self, proxy=None, proxy_loadargs=[], uisettings=None):
        if not proxy:
            proxy = ConfigurationProxy(loadargs=proxy_loadargs)
        self._current_state = VirginState(proxy, uisettings)

    def get_state(self):
        """
        Expose the underlying L{ConfigurationState}, for testing purposes.
        """
        return self._current_state

    def load_data(self, asynchronous=True, exit_method=None):
        self._current_state = self._current_state.load_data(
            asynchronous=asynchronous, exit_method=exit_method)
        return isinstance(self._current_state, InitialisedState)

    def modify(self):
        self._current_state = self._current_state.modify()

    def revert(self):
        self._current_state = self._current_state.revert()

    def persist(self):
        self._current_state = self._current_state.persist()

    def _get_management_type(self):
        return self._current_state.get(MANAGEMENT_TYPE)

    def _set_management_type(self, value):
        self._current_state.set(MANAGEMENT_TYPE, value)

    management_type = property(_get_management_type, _set_management_type)

    def _get_computer_title(self):
        return self._current_state.get(COMPUTER_TITLE)

    def _set_computer_title(self, value):
        self._current_state.set(COMPUTER_TITLE, value)

    computer_title = property(_get_computer_title, _set_computer_title)

    def _get_hosted_landscape_host(self):
        return self._current_state.get(HOSTED, LANDSCAPE_HOST)

    hosted_landscape_host = property(_get_hosted_landscape_host)

    def _get_local_landscape_host(self):
        return self._current_state.get(LOCAL, LANDSCAPE_HOST)

    def _set_local_landscape_host(self, value):
        self._current_state.set(LOCAL, LANDSCAPE_HOST, value)

    local_landscape_host = property(_get_local_landscape_host,
                                    _set_local_landscape_host)

    def _get_hosted_account_name(self):
        return self._current_state.get(HOSTED, ACCOUNT_NAME)

    def _set_hosted_account_name(self, value):
        self._current_state.set(HOSTED, ACCOUNT_NAME, value)

    hosted_account_name = property(_get_hosted_account_name,
                                   _set_hosted_account_name)

    def _get_local_account_name(self):
        return self._current_state.get(LOCAL, ACCOUNT_NAME)

    def _set_local_account_name(self, value):
        self._current_state.set(LOCAL, ACCOUNT_NAME, value)

    local_account_name = property(_get_local_account_name,
                                   _set_local_account_name)

    def _get_hosted_password(self):
        return self._current_state.get(HOSTED, PASSWORD)

    def _set_hosted_password(self, value):
        self._current_state.set(HOSTED, PASSWORD, value)

    hosted_password = property(_get_hosted_password,
                               _set_hosted_password)

    def _get_local_password(self):
        return self._current_state.get(LOCAL, PASSWORD)

    def _set_local_password(self, value):
        self._current_state.set(LOCAL, PASSWORD, value)

    local_password = property(_get_local_password,
                              _set_local_password)

    def _get_is_modified(self):
        return isinstance(self.get_state(), ModifiedState)

    is_modified = property(_get_is_modified)

    def get_config_filename(self):
        return self._current_state.get_config_filename()

    def exit(self, asynchronous=True):
        self._current_state.exit(asynchronous=asynchronous)



"""See L{WatchDog}.

The WatchDog must run as root, because it spawns the Landscape Manager.

The main C{landscape-client} program uses this watchdog.
"""

import os
import errno
import sys
import pwd
import signal
import time

from logging import warning, info, error
from resource import setrlimit, RLIMIT_NOFILE

from twisted.internet import reactor
from twisted.internet.defer import Deferred, succeed
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.error import ProcessExitedAlready
from twisted.application.service import Service, Application
from twisted.application.app import startApplication

from landscape.deployment import init_logging, Configuration
from landscape.lib.twisted_util import gather_results
from landscape.lib.log import log_failure
from landscape.lib.bootstrap import (BootstrapList, BootstrapFile,
                                     BootstrapDirectory)
from landscape.log import rotate_logs
from landscape.broker.amp import (
    RemoteBrokerConnector, RemoteMonitorConnector, RemoteManagerConnector)
from landscape.reactor import LandscapeReactor
from landscape.lib.dns import discover_server
from landscape.configuration import (
    fetch_base64_ssl_public_certificate, decode_base64_ssl_public_certificate)

GRACEFUL_WAIT_PERIOD = 10
MAXIMUM_CONSECUTIVE_RESTARTS = 5
RESTART_BURST_DELAY = 30  # seconds
SIGKILL_DELAY = 10


class DaemonError(Exception):
    """One of the daemons could not be started."""


class TimeoutError(Exception):
    """Something took too long."""


class ExecutableNotFoundError(Exception):
    """An executable was not found."""


class Daemon(object):
    """A Landscape daemon which can be started and tracked.

    This class should be subclassed to specify individual daemon.

    @cvar program: The name of the executable program that will start this
        daemon.
    @cvar username: The name of the user to switch to, by default.
    @cvar service: The DBus service name that the program will be expected to
        listen on.
    @cvar max_retries: The maximum number of retries before giving up when
        trying to connect to the watched daemon.
    @cvar factor: The factor by which the delay between subsequent connection
        attempts will increase.

    @param connector: The L{ComponentConnector} of the daemon.
    @param reactor: The reactor used to spawn the process and schedule timed
        calls.
    @param verbose: Optionally, report more information when running this
        program.  Defaults to False.
    """

    username = "landscape"
    max_retries = 3
    factor = 1.1
    options = None

    def __init__(self, connector, reactor=reactor, verbose=False,
                 config=None):
        self._connector = connector
        self._reactor = reactor
        self._env = os.environ.copy()
        my_uid = os.getuid()
        if my_uid == 0:
            pwd_info = pwd.getpwnam(self.username)
            target_uid = pwd_info.pw_uid
            target_gid = pwd_info.pw_gid
            if target_uid != my_uid:
                self._uid = target_uid
            else:
                self._uid = None
            if target_gid != os.getgid():
                self._gid = target_gid
            else:
                self._gid = None
            self._env["HOME"] = pwd_info.pw_dir
            self._env["USER"] = self.username
            self._env["LOGNAME"] = self.username
        else:
            # We can only switch UIDs if we're root, so simply don't switch
            # UIDs if we're not.
            self._uid = None
            self._gid = None
        self._verbose = verbose
        self._config = config
        self._process = None
        self._last_started = 0
        self._quick_starts = 0
        self._allow_restart = True

    def find_executable(self):
        """Find the fully-qualified path to the executable.

        If the executable can't be found, L{ExecutableNotFoundError} will be
        raised.
        """
        dirname = os.path.dirname(os.path.abspath(sys.argv[0]))
        executable = os.path.join(dirname, self.program)
        if not os.path.exists(executable):
            raise ExecutableNotFoundError("%s doesn't exist" % (executable,))
        return executable

    def start(self):
        """Start this daemon."""
        self._process = None

        now = time.time()
        if self._last_started + RESTART_BURST_DELAY > now:
            self._quick_starts += 1
            if self._quick_starts == MAXIMUM_CONSECUTIVE_RESTARTS:
                error("Can't keep %s running. Exiting." % self.program)
                self._reactor.stop()
                return
        else:
            self._quick_starts = 0

        self._last_started = now

        self._process = WatchedProcessProtocol(self)
        exe = self.find_executable()
        args = [exe, "--ignore-sigint"]
        if not self._verbose:
            args.append("--quiet")
        if self._config:
            args.extend(["-c", self._config])
        if self.options is not None:
            args.extend(self.options)
        self._reactor.spawnProcess(self._process, exe, args=args,
                                   env=self._env, uid=self._uid, gid=self._gid)

    def stop(self):
        """Stop this daemon."""
        if not self._process:
            return succeed(None)
        return self._process.kill()

    def _connect_and_call(self, name, *args, **kwargs):
        """Connect to the remote daemon over AMP and perform the given command.

        @param name: The name of the command to perform.
        @param args: Arguments list to be passed to the connect method
        @param kwargs: Keywords arguments to pass to the connect method.
        @return: A L{Deferred} resulting in C{True} if the command was
            successful or C{False} otherwise.
        @see: L{RemoteLandscapeComponentCreator.connect}.
        """

        def disconnect(ignored):
            self._connector.disconnect()
            return True

        connected = self._connector.connect(self.max_retries, self.factor,
                                            quiet=True)
        connected.addCallback(lambda remote: getattr(remote, name)())
        connected.addCallback(disconnect)
        connected.addErrback(lambda x: False)
        return connected

    def request_exit(self):
        return self._connect_and_call("exit")

    def is_running(self):
        # FIXME Error cases may not be handled in the best possible way
        # here. We're basically return False if any error happens from the
        # dbus ping.
        return self._connect_and_call("ping")

    def wait(self):
        """
        Return a Deferred which will fire when the process has died.
        """
        if not self._process:
            return succeed(None)
        return self._process.wait()

    def wait_or_die(self):
        """
        Wait for the process to die for C{GRACEFUL_WAIT_PERIOD}. If it hasn't
        died by that point, send it a SIGTERM. If it doesn't die for
        C{SIGKILL_DELAY},
        """
        if not self._process:
            return succeed(None)
        return self._process.wait_or_die()

    def prepare_for_shutdown(self):
        """Called by the watchdog when starting to shut us down.

        It will prevent our L{WatchedProcessProtocol} to restart the process
        when it exits.
        """
        self._allow_restart = False

    def allow_restart(self):
        """Return a boolean indicating if the daemon should be restarted."""
        return self._allow_restart

    def rotate_logs(self):
        self._process.rotate_logs()


class Broker(Daemon):
    program = "landscape-broker"


class Monitor(Daemon):
    program = "landscape-monitor"


class Manager(Daemon):
    program = "landscape-manager"
    username = "root"


class WatchedProcessProtocol(ProcessProtocol):
    """
    A process-watching protocol which sends any of its output to the log file
    and restarts it when it dies.
    """

    _killed = False

    def __init__(self, daemon):
        self.daemon = daemon
        self._wait_result = None
        self._delayed_really_kill = None
        self._delayed_terminate = None

    def kill(self):
        self._terminate()
        return self.wait()

    def _terminate(self, warn=False):
        if self.transport is not None:
            if warn:
                warning("%s didn't exit. Sending SIGTERM"
                        % (self.daemon.program,))
            try:
                self.transport.signalProcess(signal.SIGTERM)
            except ProcessExitedAlready:
                pass
            else:
                # Give some time for the process, and then show who's the boss.
                delayed = reactor.callLater(SIGKILL_DELAY, self._really_kill)
                self._delayed_really_kill = delayed

    def _really_kill(self):
        try:
            self.transport.signalProcess(signal.SIGKILL)
        except ProcessExitedAlready:
            pass
        else:
            warning("%s didn't die.  Sending SIGKILL." % self.daemon.program)
        self._delayed_really_kill = None

    def rotate_logs(self):
        if self.transport is not None:
            try:
                self.transport.signalProcess(signal.SIGUSR1)
            except ProcessExitedAlready:
                pass

    def wait(self):
        if self.transport.pid is None:
            return succeed(None)
        self._wait_result = Deferred()
        return self._wait_result

    def wait_or_die(self):
        self._delayed_terminate = reactor.callLater(GRACEFUL_WAIT_PERIOD,
                                                    self._terminate, warn=True)
        return self.wait()

    def outReceived(self, data):
        # it's *probably* going to always be line buffered, by accident
        sys.stdout.write(data)

    def errReceived(self, data):
        sys.stderr.write(data)

    def processEnded(self, reason):
        """The process has ended; restart it."""
        if self._delayed_really_kill is not None:
            self._delayed_really_kill.cancel()
        if (self._delayed_terminate is not None and
                self._delayed_terminate.active()):
            self._delayed_terminate.cancel()
        if self._wait_result is not None:
            self._wait_result.callback(None)
        elif self.daemon.allow_restart():
            self.daemon.start()


class WatchDog(object):
    """
    The Landscape WatchDog starts all other landscape daemons and ensures that
    they are working.
    """

    def __init__(self, reactor=reactor, verbose=False, config=None,
                 broker=None, monitor=None, manager=None,
                 enabled_daemons=None):
        landscape_reactor = LandscapeReactor()
        if enabled_daemons is None:
            enabled_daemons = [Broker, Monitor, Manager]
        if broker is None and Broker in enabled_daemons:
            broker = Broker(
                RemoteBrokerConnector(landscape_reactor, config),
                verbose=verbose, config=config.config)
        if monitor is None and Monitor in enabled_daemons:
            monitor = Monitor(
                RemoteMonitorConnector(landscape_reactor, config),
                verbose=verbose, config=config.config)
        if manager is None and Manager in enabled_daemons:
            manager = Manager(
                RemoteManagerConnector(landscape_reactor, config),
                verbose=verbose, config=config.config)

        self.broker = broker
        self.monitor = monitor
        self.manager = manager
        self.daemons = filter(None, [self.broker, self.monitor, self.manager])
        self.reactor = reactor
        self._checking = None
        self._stopping = False
        signal.signal(
            signal.SIGUSR1,
            lambda signal, frame: reactor.callFromThread(
                self._notify_rotate_logs))
        if config is not None and config.clones > 0:
            options = ["--clones", str(config.clones),
                       "--start-clones-over", str(config.start_clones_over)]
            for daemon in self.daemons:
                daemon.options = options

        self._ping_failures = {}

    def check_running(self):
        """Return a list of any daemons that are already running."""
        results = []
        for daemon in self.daemons:
            # This method is called on startup, we basically try to connect
            # a few times in fast sequence (with exponential backoff), if we
            # don't get a response we assume the daemon is not running.
            result = daemon.is_running()
            result.addCallback(lambda is_running, d=daemon: (is_running, d))
            results.append(result)

        def got_all_results(r):
            return [x[1] for x in r if x[0]]
        return gather_results(results).addCallback(got_all_results)

    def start(self):
        """
        Start all daemons. The broker will be started first, and no other
        daemons will be started before it is running and responding to DBUS
        messages.

        @return: A deferred which fires when all services have successfully
            started. If a daemon could not be started, the deferred will fail
            with L{DaemonError}.
        """
        for daemon in self.daemons:
            daemon.start()
        self.start_monitoring()

    def start_monitoring(self):
        """Start monitoring processes which have already been started."""
        # Must wait before daemons actually start, otherwise check will
        # restart them *again*.
        self._checking = self.reactor.callLater(5, self._check)

    def _restart_if_not_running(self, is_running, daemon):
        if (not is_running) and (not self._stopping):
            warning("%s failed to respond to a ping."
                    % (daemon.program,))
            if daemon not in self._ping_failures:
                self._ping_failures[daemon] = 0
            self._ping_failures[daemon] += 1
            if self._ping_failures[daemon] == 5:
                warning("%s died! Restarting." % (daemon.program,))
                stopping = daemon.stop()

                def stopped(ignored):
                    daemon.start()
                    self._ping_failures[daemon] = 0
                stopping.addBoth(stopped)
                return stopping
        else:
            self._ping_failures[daemon] = 0

    def _check(self):
        all_running = []
        for daemon in self.daemons:
            is_running = daemon.is_running()
            is_running.addCallback(self._restart_if_not_running, daemon)
            all_running.append(is_running)

        def reschedule(ignored):
            self._checking = self.reactor.callLater(5, self._check)
        gather_results(all_running).addBoth(reschedule)

    def request_exit(self):
        if self._checking is not None and self._checking.active():
            self._checking.cancel()
        # Set a flag so that the pinger will avoid restarting the daemons if a
        # ping has already been sent but not yet responded to.
        self._stopping = True

        # This tells the daemons to not automatically restart when they end
        for daemon in self.daemons:
            daemon.prepare_for_shutdown()

        def terminate_processes(broker_stopped):
            if broker_stopped:
                results = [daemon.wait_or_die() for daemon in self.daemons]
            else:
                # If request_exit fails, we should just kill the daemons
                # immediately.
                error("Couldn't request that broker gracefully shut down; "
                      "killing forcefully.")
                results = [x.stop() for x in self.daemons]
            return gather_results(results)

        result = self.broker.request_exit()
        return result.addCallback(terminate_processes)

    def _notify_rotate_logs(self):
        for daemon in self.daemons:
            daemon.rotate_logs()
        rotate_logs()


class WatchDogConfiguration(Configuration):

    def make_parser(self):
        parser = super(WatchDogConfiguration, self).make_parser()
        parser.add_option("--daemon", action="store_true",
                          help="Fork and run in the background.")
        parser.add_option("--pid-file", type="str",
                          help="The file to write the PID to.")
        parser.add_option("--monitor-only", action="store_true",
                          help="Don't enable management features. This is "
                          "useful if you want to run the client as a non-root "
                          "user.")
        return parser

    def get_enabled_daemons(self):
        daemons = [Broker, Monitor]
        if not self.monitor_only:
            daemons.append(Manager)
        return daemons


def daemonize():
    # See http://www.steve.org.uk/Reference/Unix/faq_2.html#SEC16
    if os.fork():  # launch child and...
        os._exit(0)  # kill off parent
    os.setsid()
    if os.fork():  # launch child and...
        os._exit(0)  # kill off parent again.
    # some argue that this umask should be 0, but that's annoying.
    os.umask(077)
    null = os.open('/dev/null', os.O_RDWR)
    for i in range(3):
        try:
            os.dup2(null, i)
        except OSError, e:
            if e.errno != errno.EBADF:
                raise
    os.close(null)


class WatchDogService(Service):

    def __init__(self, config):
        self._config = config
        self.watchdog = WatchDog(verbose=not config.daemon,
                                 config=config,
                                 enabled_daemons=config.get_enabled_daemons())
        self.exit_code = 0

    def autodiscover(self):
        """
        Autodiscover called if config setting config.server_autodiscover is
        True. This method allows the watchdog to attempt server autodiscovery,
        fetch the discovered landscape server's custom CA certificate, and
        write both the certificate and the updated config file with the
        discovered values.
        """
        def update_config(hostname):
            if hostname is None:
                warning("Autodiscovery returned empty hostname string. "
                        "Reverting to previous settings.")
            else:
                info("Autodiscovery found landscape server at %s. "
                     "Updating configuration values." % hostname)
                self._config.server_autodiscover = False
                self._config.url = "https://%s/message-system" % hostname
                self._config.ping_url = "http://%s/ping" % hostname
                if not self._config.ssl_public_key:
                    # If we don't have a key on this system, pull it from
                    # the auto-discovered server and write it to the filesystem
                    ssl_public_key = fetch_base64_ssl_public_certificate(
                        hostname, on_info=info, on_error=warning)
                    if ssl_public_key:
                        self._config.ssl_public_key = ssl_public_key
                        decode_base64_ssl_public_certificate(self._config)
                self._config.write()
            return hostname

        def discovery_error(result):
            warning("Autodiscovery failed.  Reverting to previous settings.")

        lookup_deferred = discover_server(
            self._config.autodiscover_srv_query_string,
            self._config.autodiscover_a_query_string)
        lookup_deferred.addCallback(update_config)
        lookup_deferred.addErrback(discovery_error)
        return lookup_deferred

    def startService(self):
        Service.startService(self)
        bootstrap_list.bootstrap(data_path=self._config.data_path,
                                 log_dir=self._config.log_dir)
        if self._config.clones > 0:

            # Let clones open an appropriate number of fds
            setrlimit(RLIMIT_NOFILE, (self._config.clones * 100,
                                      self._config.clones * 200))

            # Increase the timeout of AMP's MethodCalls.
            # XXX: we should find a better way to expose this knot, and
            # not set it globally on the class
            from landscape.lib.amp import MethodCallSender
            MethodCallSender.timeout = 300

            # Create clones log and data directories
            for i in range(self._config.clones):
                suffix = "-clone-%d" % i
                bootstrap_list.bootstrap(
                    data_path=self._config.data_path + suffix,
                    log_dir=self._config.log_dir + suffix)

        result = succeed(None)
        if self._config.server_autodiscover:
            result.addCallback(lambda _: self.autodiscover())
        result.addCallback(lambda _: self.watchdog.check_running())

        def start_if_not_running(running_daemons):
            if running_daemons:
                error("ERROR: The following daemons are already running: %s"
                      % (", ".join(x.program for x in running_daemons)))
                self.exit_code = 1
                reactor.crash()  # so stopService isn't called.
                return
            self._daemonize()
            info("Watchdog watching for daemons.")
            return self.watchdog.start()

        def die(failure):
            log_failure(failure, "Unknown error occurred!")
            self.exit_code = 2
            reactor.crash()
        result.addCallback(start_if_not_running)
        result.addErrback(die)
        return result

    def _daemonize(self):
        if self._config.daemon:
            daemonize()
            if self._config.pid_file:
                stream = open(self._config.pid_file, "w")
                stream.write(str(os.getpid()))
                stream.close()

    def stopService(self):
        info("Stopping client...")
        Service.stopService(self)

        # If CTRL-C is pressed twice in a row, the second SIGINT actually
        # kills us before subprocesses die, and that makes them hang around.
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        done = self.watchdog.request_exit()
        done.addBoth(lambda r: self._remove_pid())
        return done

    def _remove_pid(self):
        pid_file = self._config.pid_file
        if pid_file is not None and os.access(pid_file, os.W_OK):
            stream = open(pid_file)
            pid = stream.read()
            stream.close()
            if pid == str(os.getpid()):
                os.unlink(pid_file)


bootstrap_list = BootstrapList([
    BootstrapDirectory("$data_path", "landscape", "root", 0755),
    BootstrapDirectory("$data_path/package", "landscape", "root", 0755),
    BootstrapDirectory(
        "$data_path/package/hash-id", "landscape", "root", 0755),
    BootstrapDirectory(
        "$data_path/package/binaries", "landscape", "root", 0755),
    BootstrapDirectory(
        "$data_path/package/upgrade-tool", "landscape", "root", 0755),
    BootstrapDirectory("$data_path/messages", "landscape", "root", 0755),
    BootstrapDirectory("$data_path/sockets", "landscape", "root", 0750),
    BootstrapDirectory(
        "$data_path/custom-graph-scripts", "landscape", "root", 0755),
    BootstrapDirectory("$log_dir", "landscape", "root", 0755),
    BootstrapFile("$data_path/package/database", "landscape", "root", 0644)])


def clean_environment():
    """Unset dangerous environment variables.

    In particular unset all variables beginning with DEBIAN_ or DEBCONF_,
    to avoid any problems when landscape-client is invoked from its
    postinst script.  Some environment variables may be set which would affect
    *other* maintainer scripts which landscape-client invokes.
    """
    for key in os.environ.keys():
        if (key.startswith(("DEBIAN_", "DEBCONF_")) or
                key in ["LANDSCAPE_ATTACHMENTS", "MAIL"]):
            del os.environ[key]


def run(args=sys.argv, reactor=None):
    """Start the watchdog.

    This is the topmost function that kicks off the Landscape client.  It
    cleans up the environment, loads the configuration, and starts the
    reactor.

    @param args: Command line arguments, including the program name as the
        first element.
    @param reactor: The reactor to use.  If none is specified, the global
        reactor is used.
    @raise SystemExit: if command line arguments are bad, or when landscape-
        client is not running as 'root' or 'landscape'.
    """
    clean_environment()

    config = WatchDogConfiguration()
    config.load(args)

    try:
        landscape_uid = pwd.getpwnam("landscape").pw_uid
    except KeyError:
        sys.exit("The 'landscape' user doesn't exist!")

    if os.getuid() not in (0, landscape_uid):
        sys.exit("landscape-client can only be run as 'root' or 'landscape'.")

    init_logging(config, "watchdog")

    application = Application("landscape-client")
    watchdog_service = WatchDogService(config)
    watchdog_service.setServiceParent(application)

    if reactor is None:
        from twisted.internet import reactor
    # We add a small delay to work around a Twisted bug: this method should
    # only be called when the reactor is running, but we still get a
    # PotentialZombieWarning.
    reactor.callLater(0, startApplication, application, False)
    reactor.run()
    return watchdog_service.exit_code



from pwd import struct_passwd
from grp import struct_group
import csv
import subprocess
import logging


class UserManagementError(Exception):
    """Catch all error for problems with User Management."""


class UserNotFoundError(Exception):
    """Raised when a user couldn't be found by uid/username."""


class GroupNotFoundError(Exception):
    """Raised when a group couldn't be found by gid/groupname."""


class UserProviderBase(object):
    """This is a base class for user Providers."""

    def __init__(self, locked_users=None):
        self.locked_users = locked_users or []

        self._min_uid = 1000
        self._max_uid = 60000

    def get_users(self):
        """Returns a list of all local users on the computer.

        Each user is represented as a dict with the keys: C{username},
        C{name}, C{uid}, C{enabled}, C{location}, C{work-phone} and
        C{home-phone}.  
        """
        users = []
        found_usernames = set()
        for user in self.get_user_data():
            if not isinstance(user, struct_passwd):
                user = struct_passwd(user)
            if user.pw_name in found_usernames:
                continue
            gecos_data = [x.decode("utf-8", "replace") or None
                          for x in user.pw_gecos.split(",")[:4]]
            while len(gecos_data) < 4:
                gecos_data.append(None)
            name, location, work_phone, home_phone = tuple(gecos_data)
            enabled = user.pw_name not in self.locked_users
            users.append({"username": user.pw_name, "name": name,
                          "uid": user.pw_uid, "enabled": enabled,
                          "location": location, "work-phone": work_phone,
                          "home-phone": home_phone,
                          "primary-gid": user.pw_gid})
            found_usernames.add(user.pw_name)
        return users

    def get_groups(self):
        """Returns a list of groups on the computer.

        Each group is represented as a dict with the keys: C{name},
        C{gid} and C{members}.
        """
        user_names = set([x["username"] for x in self.get_users()])
        groups = []
        found_groupnames = set()
        for group in self.get_group_data():
            if not isinstance(group, struct_group):
                group = struct_group(group)
            if group.gr_name in found_groupnames:
                continue
            member_names = user_names.intersection(group.gr_mem)
            groups.append({"name": group.gr_name, "gid": group.gr_gid,
                           "members": list(member_names)})
            found_groupnames.add(group.gr_name)
        return groups

    def get_uid(self, username):
        """Returns the UID for C{username}.

        @raises UserNotFoundError: Raised if C{username} doesn't match a
            user on the computer.
        """
        for data in self.get_users():
            if data["username"] == username:
                return data["uid"]
        raise UserNotFoundError("UID not found for user %s." % username)

    def get_gid(self, groupname):
        """Returns the GID for C{groupname}.

        @raises UserManagementError: Raised if C{groupname} doesn't
            match a group on the computer.
        """
        for data in self.get_groups():
            if data["name"] == groupname:
                return data["gid"]
        raise GroupNotFoundError("Group not found for group %s." % groupname)

class UserProvider(UserProviderBase):

    popen = subprocess.Popen

    passwd_fields = ["username", "passwd", "uid", "primary-gid", "gecos",
                     "home", "shell"]
    group_fields = ["name", "passwd", "gid", "members"]

    def __init__(self, locked_users=[], passwd_file="/etc/passwd",
                 group_file="/etc/group"):
        super(UserProvider, self).__init__(locked_users)
        self._passwd_file = passwd_file
        self._group_file = group_file

    def get_user_data(self):
        """
        Parse passwd(5) formatted files and return tuples of user data in the
        form (username, password, uid, primary-group-id, gecos data, home
        directory, path to the user's shell)
        """
        user_data = []
        passwd_file = open(self._passwd_file, "r")
        reader = csv.DictReader(passwd_file, fieldnames=self.passwd_fields,
                                delimiter=":", quoting=csv.QUOTE_NONE)
        current_line = 0
        for row in reader:
            current_line += 1
            # This skips the NIS user marker in the passwd file.
            if (row["username"].startswith("+") or
                row["username"].startswith("-")):
                continue
            try:
                user_data.append((row["username"], row["passwd"],
                                  int(row["uid"]), int(row["primary-gid"]),
                                  row["gecos"], row["home"], row["shell"]))
            except (ValueError, TypeError):

                logging.warn("passwd file %s is incorrectly formatted: "
                             "line %d." % (self._passwd_file, current_line))

        passwd_file.close()
        return user_data

    def get_group_data(self):
        """
        Parse group(5) formatted files and return tuples of group data in the
        form (groupname, group password, group id and a list of member
        usernames).
        """
        group_data = []
        group_file = open(self._group_file, "r")
        reader = csv.DictReader(group_file, fieldnames=self.group_fields,
                                delimiter=":", quoting=csv.QUOTE_NONE)
        current_line = 0
        for row in reader:
            current_line += 1
            # Skip if we find the NIS marker
            if (row["name"].startswith("+") or
                row["name"].startswith("-")):
                continue
            try:
                group_data.append((row["name"], row["passwd"], int(row["gid"]), 
                                   row["members"].split(",")))
            except (AttributeError, ValueError):
                logging.warn("group file %s is incorrectly formatted: "
                             "line %d." % (self._group_file, current_line))
        group_file.close()
        return group_data
 



import os
import time

from landscape.accumulate import Accumulator
from landscape.lib.monitor import CoverageMonitor
from landscape.monitor.plugin import MonitorPlugin


class LoadAverage(MonitorPlugin):
    """Plugin captures information about load average."""

    persist_name = "load-average"
    scope = "load"
    # Prevent the Plugin base-class from scheduling looping calls.
    run_interval = None

    def __init__(self, interval=15, monitor_interval=60*60,
                 create_time=time.time, get_load_average=os.getloadavg):
        self._interval = interval
        self._monitor_interval = monitor_interval
        self._create_time = create_time
        self._load_averages = []
        self._get_load_average = get_load_average

    def register(self, registry):
        super(LoadAverage, self).register(registry)
        self._accumulate = Accumulator(self._persist, registry.step_size)

        self.registry.reactor.call_every(self._interval, self.run)

        self._monitor = CoverageMonitor(self._interval, 0.8,
                                        "load average snapshot",
                                        create_time=self._create_time)
        self.registry.reactor.call_every(self._monitor_interval,
                                         self._monitor.log)
        self.registry.reactor.call_on("stop", self._monitor.log, priority=2000)
        self.call_on_accepted("load-average", self.send_message, True)

    def create_message(self):
        load_averages = self._load_averages
        self._load_averages = []
        return {"type": "load-average", "load-averages": load_averages}

    def exchange(self, urgent=False):
        self.registry.broker.call_if_accepted("load-average",
                                              self.send_message, urgent)

    def send_message(self, urgent=False):
        message = self.create_message()
        if len(message["load-averages"]):
            self.registry.broker.send_message(message, self._session_id,
                                              urgent=urgent)

    def run(self):
        self._monitor.ping()
        new_timestamp = int(self._create_time())
        new_load_average = self._get_load_average()[0]
        step_data = self._accumulate(new_timestamp, new_load_average,
                                     "accumulate")
        if step_data:
            self._load_averages.append(step_data)



from twisted.internet.defer import DeferredList, Deferred
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.process import Process, ProcessReader
from twisted.internet import reactor

import cStringIO


def gather_results(deferreds, consume_errors=False):
    d = DeferredList(deferreds, fireOnOneErrback=1,
                     consumeErrors=consume_errors)
    d.addCallback(lambda r: [x[1] for x in r])
    d.addErrback(lambda f: f.value.subFailure)
    return d


class AllOutputProcessProtocol(ProcessProtocol):
    """A process protocol for getting stdout, stderr and exit code."""

    def __init__(self, deferred, stdin=None, line_received=None):
        self.deferred = deferred
        self.outBuf = cStringIO.StringIO()
        self.errBuf = cStringIO.StringIO()
        self.errReceived = self.errBuf.write
        self.stdin = stdin
        self.line_received = line_received
        self._partial_line = ""

    def connectionMade(self):
        if self.stdin is not None:
            self.transport.write(self.stdin)
            self.transport.closeStdin()

    def outReceived(self, data):
        self.outBuf.write(data)

        if self.line_received is None:
            return

        # data may contain more than one line, so we split the output and save
        # the last line. If it's an empty string nothing happens, otherwise it
        # will be returned once complete
        lines = data.split("\n")
        lines[0] = self._partial_line + lines[0]
        self._partial_line = lines.pop()

        for line in lines:
            self.line_received(line)

    def processEnded(self, reason):
        if self._partial_line:
            self.line_received(self._partial_line)
            self._partial_line = ""
        out = self.outBuf.getvalue()
        err = self.errBuf.getvalue()
        e = reason.value
        code = e.exitCode
        if e.signal:
            self.deferred.errback((out, err, e.signal))
        else:
            self.deferred.callback((out, err, code))


def spawn_process(executable, args=(), env={}, path=None, uid=None, gid=None,
                  usePTY=False, wait_pipes=True, line_received=None,
                  stdin=None):
    """
    Spawn a process using Twisted reactor.

    Return a deferred which will be called with process stdout, stderr and exit
    code.

    @param wait_pipes: if set to False, don't wait for stdin/stdout pipes to
        close when process ends.
    @param line_received: an optional callback called with every line of
        output from the process as parameter.

    @note: compared to reactor.spawnProcess, this version does NOT require the
    executable name as first element of args.
    """

    list_args = [executable]
    list_args.extend(args)

    result = Deferred()
    protocol = AllOutputProcessProtocol(result, stdin=stdin,
                                        line_received=line_received)
    process = reactor.spawnProcess(protocol, executable, args=list_args,
                                   env=env, path=path, uid=uid, gid=gid,
                                   usePTY=usePTY)

    if not wait_pipes:

        def maybeCallProcessEnded():
            """A less strict version of Process.maybeCallProcessEnded.

            This behaves exactly like the original method, but in case the
            process has ended already and sent us a SIGCHLD, it doesn't wait
            for the stdin/stdout pipes to close, because the child process
            itself might have passed them to its own child processes.

            @note: Twisted 8.2 now has a processExited hook that could
                be used in place of this workaround.
            """
            if process.pipes and not process.pid:
                for pipe in process.pipes.itervalues():
                    if isinstance(pipe, ProcessReader):
                        # Read whatever is left
                        pipe.doRead()
                    pipe.stopReading()
                process.pipes = {}
            Process.maybeCallProcessEnded(process)

        process.maybeCallProcessEnded = maybeCallProcessEnded

    return result



import logging

from landscape.lib.encoding import encode_dict_if_needed
from landscape.amp import ComponentConnector, ComponentPublisher, remote

from landscape.user.management import UserManagement
from landscape.manager.plugin import ManagerPlugin
from landscape.monitor.usermonitor import RemoteUserMonitorConnector


class UserManager(ManagerPlugin):

    name = "usermanager"

    def __init__(self, management=None, shadow_file="/etc/shadow"):
        self._management = management or UserManagement()
        self._shadow_file = shadow_file
        self._message_types = {"add-user": self._add_user,
                               "edit-user": self._edit_user,
                               "lock-user": self._lock_user,
                               "unlock-user": self._unlock_user,
                               "remove-user": self._remove_user,
                               "add-group": self._add_group,
                               "edit-group": self._edit_group,
                               "remove-group": self._remove_group,
                               "add-group-member": self._add_group_member,
                               "remove-group-member":
                               self._remove_group_member}
        self._publisher = None

    def register(self, registry):
        """
        Schedule reactor events for generic L{Plugin} callbacks, user
        and group management operations, and resynchronization.
        """
        super(UserManager, self).register(registry)
        self._registry = registry

        self._publisher = ComponentPublisher(self, self.registry.reactor,
                                             self.registry.config)
        self._publisher.start()

        for message_type in self._message_types:
            self._registry.register_message(message_type,
                                            self._message_dispatch)

    def stop(self):
        """Stop listening for incoming AMP connections."""
        if self._publisher:
            self._publisher.stop()
            self._publisher = None

    @remote
    def get_locked_usernames(self):
        """Return a list of usernames with locked system accounts."""
        locked_users = []
        if self._shadow_file:
            try:
                shadow_file = open(self._shadow_file, "r")
                for line in shadow_file:
                    parts = line.split(":")
                    if len(parts) > 1:
                        if parts[1].startswith("!"):
                            locked_users.append(parts[0].strip())
            except IOError, e:
                logging.error("Error reading shadow file. %s" % e)
        return locked_users

    def _message_dispatch(self, message):
        """Dispatch the given user-change request to the correct handler.

        @param message: The request we got from the server.
        """
        user_monitor_connector = RemoteUserMonitorConnector(
            self.registry.reactor, self.registry.config)

        def detect_changes(user_monitor):
            self._user_monitor = user_monitor
            return user_monitor.detect_changes()

        result = user_monitor_connector.connect()
        result.addCallback(detect_changes)
        result.addCallback(self._perform_operation, message)
        result.addCallback(self._send_changes, message)
        result.addCallback(lambda x: user_monitor_connector.disconnect())
        return result

    def _perform_operation(self, result, message):
        message_type = message["type"]
        message_method = self._message_types[message_type]
        message = encode_dict_if_needed(message)
        return self.call_with_operation_result(message, message_method,
                                               message)

    def _send_changes(self, result, message):
        return self._user_monitor.detect_changes(message["operation-id"])

    def _add_user(self, message):
        """Run an C{add-user} operation."""
        return self._management.add_user(message["username"], message["name"],
                                         message["password"],
                                         message["require-password-reset"],
                                         message["primary-group-name"],
                                         message["location"],
                                         message["work-number"],
                                         message["home-number"])

    def _edit_user(self, message):
        """Run an C{edit-user} operation."""
        return self._management.set_user_details(
                 message["username"], password=message["password"],
                 name=message["name"], location=message["location"],
                 work_number=message["work-number"],
                 home_number=message["home-number"],
                 primary_group_name=message["primary-group-name"])

    def _lock_user(self, message):
        """Run a C{lock-user} operation."""
        return self._management.lock_user(message["username"])

    def _unlock_user(self, message):
        """Run an C{unlock-user} operation."""
        return self._management.unlock_user(message["username"])

    def _remove_user(self, message):
        """Run a C{remove-user} operation."""
        return self._management.remove_user(message["username"],
                                            message["delete-home"])

    def _add_group(self, message):
        """Run an C{add-group} operation."""
        return self._management.add_group(message["groupname"])

    def _edit_group(self, message):
        """Run an C{edit-group} operation."""
        return self._management.set_group_details(message["groupname"],
                                                  message["new-name"])

    def _add_group_member(self, message):
        """Run an C{add-group-member} operation."""
        return self._management.add_group_member(message["username"],
                                                 message["groupname"])

    def _remove_group_member(self, message):
        """Run a C{remove-group-member} operation."""
        return self._management.remove_group_member(message["username"],
                                                    message["groupname"])

    def _remove_group(self, message):
        """Run an C{remove-group} operation."""
        return self._management.remove_group(message["groupname"])


class RemoteUserManagerConnector(ComponentConnector):

    component = UserManager



from landscape.lib.persist import Persist
from landscape.accumulate import Accumulator, accumulate
from landscape.tests.helpers import LandscapeTest


class AccumulateTest(LandscapeTest):
    """Test for the accumulate function that implements accumulation logic."""

    def test_accumulate(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   0              4
        """
        accumulated_value, step_data = accumulate(0, 0, 5, 4, 5)
        self.assertEqual(accumulated_value, 0)
        self.assertEqual(step_data, (5, 4))

    def test_accumulate_non_zero_accumulated_value(self):
        """
        step:    5              10             15
               --|--+--+--+--+--|--+--+--+--+--|--
        value:         4                 3
        """
        accumulated_value, step_data = accumulate(7, 8, 13, 3, 5)
        self.assertEqual(accumulated_value, 9)
        self.assertEqual(step_data, (10, float((2 * 4) + (3 * 3)) / 5))

    def test_accumulate_skipped_step(self):
        """
        step:    0              5              10             15
               --|--+--+--+--+--|--+--+--+--+--|--+--+--+--+--|--
        value:   0                                   4
        """
        accumulated_value, step_data = accumulate(0, 0, 12, 4, 5)
        self.assertEqual(accumulated_value, 8)
        self.assertEqual(step_data, None)

    def test_accumulate_within_step(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   0     4
        """
        accumulated_value, step_data = accumulate(0, 0, 2, 4, 5)
        self.assertEqual(accumulated_value, 8)
        self.assertEqual(step_data, None)

    def test_accumulate_within_step_with_nonzero_start_accumulated_value(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   0     3     4
        """
        accumulated_value, step_data = accumulate(2, 6, 4, 4, 5)
        self.assertEqual(accumulated_value, 14)
        self.assertEqual(step_data, None)

    def test_accumulate_with_first_value_on_step_boundary(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   14
        """
        accumulated_value, step_data = accumulate(0, 0, 0, 14, 5)
        self.assertEqual(accumulated_value, 0)
        self.assertEqual(step_data, None)


class AccumulatorTest(LandscapeTest):
    """Tests for the Accumulator plugin helper class."""

    def test_accumulate(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   0              4
        """
        persist = Persist()
        accumulate = Accumulator(persist, 5)

        self.assertEqual(persist.get("key"), None)
        step_data = accumulate(5, 4, "key")
        self.assertEqual(step_data, (5, 4))
        self.assertEqual(persist.get("key"), (5, 0))

    def test_accumulate_non_zero_accumulated_value(self):
        """
        step:    5              10             15
               --|--+--+--+--+--|--+--+--+--+--|--
        value:         4                 3
        """
        persist = Persist()
        accumulate = Accumulator(persist, 5)

        # Persist data that would have been stored when
        # accumulate(7, 4, "key") was called.
        persist.set("key", (7, 8))
        step_data = accumulate(13, 3, "key")
        self.assertEqual(step_data, (10, float((2 * 4) + (3 * 3)) / 5))
        self.assertEqual(persist.get("key"), (13, 9))

    def test_accumulate_skipped_step(self):
        """
        step:    0              5              10             15
               --|--+--+--+--+--|--+--+--+--+--|--+--+--+--+--|--
        value:   0                                   4
        """
        persist = Persist()
        accumulate = Accumulator(persist, 5)

        self.assertEqual(persist.get("key"), None)
        step_data = accumulate(12, 4, "key")
        self.assertEqual(step_data, None)
        self.assertEqual(persist.get("key"), (12, 8))

    def test_accumulate_within_step(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   0     4
        """
        persist = Persist()
        accumulate = Accumulator(persist, 5)

        self.assertEqual(persist.get("key"), None)
        step_data = accumulate(2, 4, "key")
        self.assertEqual(step_data, None)
        self.assertEqual(persist.get("key"), (2, 8))

    def test_accumulate_within_step_with_nonzero_start_accumulated_value(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   0     3     4
        """
        persist = Persist()
        accumulate = Accumulator(persist, 5)

        # Persist data that would have been stored when
        # accumulate(2, 3, "key") was called.
        persist.set("key", (2, 6))
        step_data = accumulate(4, 4, "key")
        self.assertEqual(step_data, None)
        self.assertEqual(persist.get("key"), (4, 14))

    def test_accumulate_with_first_value_on_step_boundary(self):
        """
        step:    0              5
               --|--+--+--+--+--|--
        value:   14
        """
        persist = Persist()
        accumulate = Accumulator(persist, 5)

        self.assertEqual(persist.get("key"), None)
        step_data = accumulate(0, 14, "key")
        self.assertEqual(step_data, None)
        self.assertEqual(persist.get("key"), (0, 0))




from landscape.tests.helpers import LandscapeTest
from landscape.ui.model.configuration.uisettings import UISettings
from landscape.ui.tests.helpers import (
    FakeGSettings, dbus_test_should_skip, dbus_skip_message)


class UISettingsTest(LandscapeTest):

    default_data = {"management-type": "LDS",
                    "computer-title": "bound.to.lose",
                    "hosted-landscape-host": "landscape.canonical.com",
                    "hosted-account-name": "Sparklehorse",
                    "hosted-password": "Vivadixiesubmarinetransmissionplot",
                    "local-landscape-host": "the.local.machine",
                    "local-account-name": "CrazyHorse",
                    "local-password": "RustNeverSleeps"}

    def setUp(self):
        super(UISettingsTest, self).setUp()
        self.settings = FakeGSettings(data=self.default_data)
        self.uisettings = UISettings(self.settings)

    def test_setup(self):
        """
        Test that the L{GSettings.Client} is correctly initialised.
        """
        self.assertTrue(self.settings.was_called_with_args(
                "new", UISettings.BASE_KEY))

    def test_get_management_type(self):
        """
        Test that the L{get_management_type} value is correctly fetched from
        the L{GSettings.Client}.
        """
        self.assertEqual("LDS", self.uisettings.get_management_type())

    def test_set_management_type(self):
        """
        Test that we can correctly use L{set_management_type} to write the
        L{management_type} value to the L{GSettings.Client}.
        """
        self.assertEqual("LDS", self.uisettings.get_management_type())
        self.uisettings.set_management_type("canonical")
        self.assertEqual("canonical", self.uisettings.get_management_type())
        self.assertTrue(self.settings.was_called_with_args(
                "set_string", "management-type", "canonical"))

    def test_get_computer_title(self):
        """
        Test that the L{get_computer_title} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("bound.to.lose",
                         self.uisettings.get_computer_title())

    def test_set_computer_title(self):
        """
        Test that L{set_computer_title} correctly sets the value of
        L{computer_title} in the L{GSettings.Client}.
        """
        self.assertEqual("bound.to.lose", self.uisettings.get_computer_title())
        self.uisettings.set_computer_title("Bang")
        self.assertEqual("Bang", self.uisettings.get_computer_title())

    def test_get_hosted_landscape_host(self):
        """
        Test that the L{get_hosted_landscape_host} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("landscape.canonical.com",
                         self.uisettings.get_hosted_landscape_host())

    # NOTE: There is no facility to set the hosted-landscape-host

    def test_get_hosted_account_name(self):
        """
        Test that the L{get_hosted_account_name} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("Sparklehorse",
                         self.uisettings.get_hosted_account_name())

    def test_set_hosted_account_name(self):
        """
        Test that L{set_hosted_account_name} correctly sets the value of
        L{hosted_account_name} in the L{GSettings.Client}.
        """
        self.assertEqual("Sparklehorse",
                         self.uisettings.get_hosted_account_name())
        self.uisettings.set_hosted_account_name("Bang")
        self.assertEqual("Bang", self.uisettings.get_hosted_account_name())

    def test_get_hosted_password(self):
        """
        Test that the L{get_hosted_password} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("Vivadixiesubmarinetransmissionplot",
                         self.uisettings.get_hosted_password())

    def test_set_hosted_password(self):
        """
        Test that L{set_hosted_password} correctly sets the value of
        L{hosted_password} in the L{GSettings.Client}.
        """
        self.assertEqual("Vivadixiesubmarinetransmissionplot",
                         self.uisettings.get_hosted_password())
        self.uisettings.set_hosted_password("Bang")
        self.assertEqual("Bang", self.uisettings.get_hosted_password())

    def test_get_local_landscape_host(self):
        """
        Test that the L{get_local_landscape_host} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("the.local.machine",
                         self.uisettings.get_local_landscape_host())

    def test_set_local_landscape_host(self):
        """
        Test that L{set_local_landscape_host} correctly sets the value of
        L{local_landscape_host} in the L{GSettings.Client}.
        """
        self.assertEqual("the.local.machine",
                         self.uisettings.get_local_landscape_host())
        self.uisettings.set_local_landscape_host("Bang")
        self.assertEqual("Bang", self.uisettings.get_local_landscape_host())

    def test_get_local_account_name(self):
        """
        Test that the L{get_local_account_name} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("CrazyHorse",
                         self.uisettings.get_local_account_name())

    def test_set_local_account_name(self):
        """
        Test that L{set_local_account_name} correctly sets the value of
        L{local_account_name} in the L{GSettings.Client}.
        """
        self.assertEqual("CrazyHorse",
                         self.uisettings.get_local_account_name())
        self.uisettings.set_local_account_name("Bang")
        self.assertEqual("Bang", self.uisettings.get_local_account_name())

    def test_get_local_password(self):
        """
        Test that the L{get_local_password} value is correctly fetched
        from the L{GSettings.Client}.
        """
        self.assertEqual("RustNeverSleeps",
                         self.uisettings.get_local_password())

    def test_set_local_password(self):
        """
        Test that L{set_local_password} correctly sets the value of
        L{local_password} in the L{GSettings.Client}.
        """
        self.assertEqual("RustNeverSleeps",
                         self.uisettings.get_local_password())
        self.uisettings.set_local_password("Bang")
        self.assertEqual("Bang", self.uisettings.get_local_password())

    if dbus_test_should_skip:
        skip = dbus_skip_message
